#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Tests tools
"""

from dessia_common.forms import StandaloneSubobject, EmbeddedSubobject, EnhancedStandaloneSubobject, StandaloneObject,\
    InheritingStandaloneSubobject

standalone_subobject = StandaloneSubobject(floatarg=3.78)
embedded_subobject = EmbeddedSubobject()
object_list = [StandaloneSubobject(floatarg=666.999), StandaloneSubobject(floatarg=999.666)]
subobject_list = [EmbeddedSubobject(), EmbeddedSubobject(), EmbeddedSubobject(), EmbeddedSubobject()]
union_object = EnhancedStandaloneSubobject(floatarg=333.333, boolarg=True)

subclass_arg = InheritingStandaloneSubobject(floatarg=1561.57, strarg='Subclass')

standalone_object = StandaloneObject(standalone_subobject=standalone_subobject, embedded_subobject=embedded_subobject,
                                     dynamic_dict={'key0': True, 'key1': False}, float_dict={'a': 1.3, 'b': 3.1},
                                     string_dict={'key0': 'value0', 'key1': 'value1'},
                                     tuple_arg=('Tuple', 1), intarg=5, strarg='TestStr', object_list=object_list,
                                     subobject_list=subobject_list, builtin_list=[1, 2, 3, 4, 5],
                                     union_arg=union_object, subclass_arg=subclass_arg,
                                     array_arg=[[1, 2, 3], [4, None, 6]])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module that stores warnings in dessia_common
"""

class SerializationWarning(Warning):
    def __init__(self, message: str):
        self.message = message

        Warning.__init__(self, message)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for errors in dessia_common
"""


class ExceptionWithTraceback(Exception):
    """
    Base class with a message and a traceback
    """

    def __init__(self, message, traceback_=''):
        super().__init__(message)
        self.message = message
        self.traceback = traceback_

    def __str__(self):
        return f'{self.message}\nTraceback:\n{self.traceback}'


class DeepAttributeError(ExceptionWithTraceback, AttributeError):
    pass


class ModelError(Exception):
    pass


class ConsistencyError(Exception):
    pass


class SerializationError(Exception):
    pass


class DeserializationError(Exception):
    pass


class CopyError(Exception):
    pass


class UntypedArgumentError(Exception):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Serialization Tools

"""

import warnings
import inspect
import collections
import collections.abc
from ast import literal_eval
from typing import get_origin, get_args, Union, Any, BinaryIO, TextIO
from numpy import int64, float64
import networkx as nx
import dessia_common as dc
import dessia_common.core
import dessia_common.errors as dc_err
from dessia_common.files import StringFile, BinaryFile
import dessia_common.utils.types as dcty
from dessia_common.typings import InstanceOf
from dessia_common.graph import explore_tree_from_leaves  # , cut_tree_final_branches
from dessia_common.breakdown import get_in_object_from_path


def serialize_dict(dict_):
    """
    Serialize a dict into a dict (values are serialized)
    """
    serialized_dict = {}
    for key, value in dict_.items():
        serialized_dict[key] = serialize(value)
    return serialized_dict


def serialize_sequence(seq):
    """
    Serialize a sequence (list or sequence) into a list of dicts
    """
    serialized_sequence = []
    for value in seq:
        serialized_sequence.append(serialize(value))
    return serialized_sequence


def serialize(value):
    """
    Main function for serialization without pointers
    Calls recursively itself serialize_sequence and serialize_dict
    """
    if isinstance(value, (dc.DessiaObject, dessia_common.core.DessiaObject)):
        try:
            serialized_value = value.to_dict(use_pointers=False)
        except TypeError:
            warnings.warn(f'specific to_dict of class {value.__class__} '
                          'should implement use_pointers, memo and path arguments', Warning)
            serialized_value = value.to_dict()
    elif isinstance(value, dict):
        serialized_value = serialize_dict(value)
    elif dcty.is_sequence(value):
        serialized_value = serialize_sequence(value)
    elif isinstance(value, (BinaryFile, StringFile)):
        serialized_value = value
    elif isinstance(value, int64):
        serialized_value = int(value)
    elif isinstance(value, float64):
        serialized_value = float(value)
    elif isinstance(value, type) or dcty.is_typing(value):
        return dcty.serialize_typing(value)
    elif hasattr(value, 'to_dict'):
        to_dict_method = getattr(value, 'to_dict', None)
        if callable(to_dict_method):
            return to_dict_method()
    else:
        if not dcty.is_jsonable(value):
            msg = f'Element of value {value} is not json serializable'
            raise dc_err.SerializationError(msg)
        serialized_value = value
    return serialized_value


def serialize_with_pointers(value, memo=None, path='#'):
    """
    Main function for serialization with pointers
    """
    if memo is None:
        memo = {}
    if isinstance(value, (dc.DessiaObject, dessia_common.core.DessiaObject)):
        if value in memo:
            return {'$ref': memo[value]}, memo
        try:
            serialized = value.to_dict(use_pointers=True, memo=memo, path=path)

        except TypeError:
            warnings.warn('specific to_dict should implement use_pointers, memo and path arguments', Warning)
            serialized = value.to_dict()
        memo[value] = path

    elif hasattr(value, 'to_dict'):
        if value in memo:
            return {'$ref': memo[value]}, memo
        serialized = value.to_dict()
        memo[value] = path
    elif isinstance(value, dict):
        serialized, memo = serialize_dict_with_pointers(value, memo=memo, path=path)
    elif dcty.is_sequence(value):
        serialized, memo = serialize_sequence_with_pointers(value, memo=memo, path=path)

    elif isinstance(value, (BinaryFile, StringFile)):
        serialized = value
    elif isinstance(value, (int64, float64)):
        serialized = serialize(value)
    else:
        if not dcty.is_jsonable(value):
            msg = f'Element of value {value} is not json serializable'
            raise dc_err.SerializationError(msg)
        serialized = value
    return serialized, memo


def serialize_dict_with_pointers(dict_, memo, path):
    """
    Serialize a dict recursively with jsonpointers using a memo dict at a given path of the top level object
    """
    serialized_dict = {}
    dict_attrs_keys = []
    seq_attrs_keys = []
    other_keys = []
    # Detecting type of keys
    for key, value in dict_.items():
        if isinstance(value, dict):
            dict_attrs_keys.append(key)
        elif dcty.is_sequence(value):
            seq_attrs_keys.append(key)
        else:
            other_keys.append(key)

    for key in other_keys:
        value_path = f'{path}/{key}'
        serialized_dict[key], memo = serialize_with_pointers(dict_[key], memo=memo, path=value_path)
    # Handle seq & dicts afterwards
    for key in seq_attrs_keys:
        value_path = f'{path}/{key}'
        serialized_dict[key], memo = serialize_sequence_with_pointers(dict_[key], memo=memo, path=value_path)

    for key in dict_attrs_keys:
        value_path = f'{path}/{key}'
        serialized_dict[key], memo = serialize_dict_with_pointers(dict_[key], memo=memo, path=value_path)
    return serialized_dict, memo


def serialize_sequence_with_pointers(seq, memo, path):
    """
    Serialize a sequence (list or tuple) using jsonpointers
    """
    serialized_sequence = []
    for ival, value in enumerate(seq):
        value_path = f'{path}/{ival}'
        serialized_value, memo = serialize_with_pointers(value, memo=memo, path=value_path)
        serialized_sequence.append(serialized_value)

    return serialized_sequence, memo


def deserialize(serialized_element, sequence_annotation: str = 'List',
                global_dict=None, pointers_memo=None, path: str = '#'):
    """
    Main function for deserialization, handle pointers
    """
    if pointers_memo is not None:
        if path in pointers_memo:
            return pointers_memo[path]

    if isinstance(serialized_element, dict):
        # try:
        return dict_to_object(serialized_element, global_dict=global_dict, pointers_memo=pointers_memo, path=path)
        # except TypeError:
        #     warnings.warn(f'specific dict_to_object of class {serialized_element.__class__.__name__}'
        #                   ' should implement global_dict and'
        #                   ' pointers_memo arguments',
        #                   Warning)
        #     return dict_to_object(serialized_element)
    if dcty.is_sequence(serialized_element):
        return deserialize_sequence(sequence=serialized_element, annotation=sequence_annotation,
                                    global_dict=global_dict, pointers_memo=pointers_memo, path=path)
    return serialized_element


def deserialize_sequence(sequence, annotation=None,
                         global_dict=None, pointers_memo=None,
                         path='#'):
    # TODO: rename to deserialize sequence? Or is this a duplicate ?
    origin, args = dcty.unfold_deep_annotation(typing_=annotation)
    deserialized_sequence = []
    for ie, elt in enumerate(sequence):
        path_elt = f'{path}/{ie}'
        deserialized_element = deserialize(elt, args, global_dict=global_dict,
                                           pointers_memo=pointers_memo, path=path_elt)
        deserialized_sequence.append(deserialized_element)
    if origin is tuple:
        # Keeping as a tuple
        return tuple(deserialized_sequence)
    return deserialized_sequence


def dict_to_object(dict_, class_=None, force_generic: bool = False,
                   global_dict=None, pointers_memo=None, path='#'):
    """
    Transform a dict to an object
    """
    class_argspec = None

    global_dict, pointers_memo = update_pointers_data(global_dict=global_dict, current_dict=dict_,
                                                      pointers_memo=pointers_memo)

    if '$ref' in dict_:
        try:
            return pointers_memo[dict_['$ref']]
        except KeyError as err:
            print('keys in memo:')
            for key in sorted(pointers_memo.keys()):
                print(f'\t{key}')
            raise RuntimeError(f"Pointer {dict_['$ref']} not in memo, at path {path}") from err

    if class_ is None and 'object_class' in dict_:
        class_ = dcty.get_python_class_from_class_name(dict_['object_class'])

    # Create init_dict
    if class_ is not None and hasattr(class_, 'dict_to_object'):
        different_methods = (class_.dict_to_object.__func__ is not dc.DessiaObject.dict_to_object.__func__)
        if different_methods and not force_generic:
            try:
                obj = class_.dict_to_object(dict_, global_dict=global_dict, pointers_memo=pointers_memo, path=path)
            except TypeError:
                warnings.warn(f'specific to_dict of class {class_.__name__} '
                              'should implement use_pointers, memo and path arguments', Warning)
                obj = class_.dict_to_object(dict_)

            return obj

        class_argspec = inspect.getfullargspec(class_)
        init_dict = {k: v for k, v in dict_.items() if k in class_argspec.args + class_argspec.kwonlyargs}
        # TOCHECK Class method to generate init_dict ??
    else:
        init_dict = dict_

    subobjects = {}
    for key, value in init_dict.items():
        if class_argspec is not None and key in class_argspec.annotations:
            annotation = class_argspec.annotations[key]
        else:
            annotation = None

        key_path = f'{path}/{key}'

        if key_path in pointers_memo:
            subobjects[key] = pointers_memo[key_path]
        else:
            subobjects[key] = deserialize(value, annotation, global_dict=global_dict,
                                          pointers_memo=pointers_memo, path=key_path)  # , enforce_pointers=False)
    if class_ is not None:
        obj = class_(**subobjects)
    else:
        obj = subobjects

    return obj


def deserialize_with_type(type_, value):
    if type_ in dcty.TYPES_STRINGS.values():
        return literal_eval(type_)(value)
    if isinstance(type_, str):
        class_ = dcty.get_python_class_from_class_name(type_)
        if inspect.isclass(class_):
            return class_.dict_to_object(value)
        raise NotImplementedError(f'Cannot get class from name {type_}')

    if isinstance(type_, (list, tuple)):
        return [deserialize_with_type(t, v) for t, v in zip(type_, value)]
    if type_ is None:
        return value

    raise NotImplementedError(type_)


def deserialize_with_typing(type_, argument):
    """
    Deserialize an object with a typing info
    """
    origin = get_origin(type_)
    args = get_args(type_)
    deserialized_arg = None
    if origin is Union:
        # Check for Union false Positive (Default value = None)
        if len(args) == 2 and type(None) in args:
            return deserialize_argument(type_=args[0], argument=argument)

        # Type union
        classes = list(args)
        instantiated = False
        while instantiated is False:
            # Find the last class in the hierarchy
            hierarchy_lengths = [len(cls.mro()) for cls in classes]
            max_length = max(hierarchy_lengths)
            children_class_index = hierarchy_lengths.index(max_length)
            children_class = classes[children_class_index]
            try:
                # Try to deserialize
                # Throws KeyError if we try to put wrong dict into
                # dict_to_object. This means we try to instantiate
                # a children class with a parent dict_to_object
                deserialized_arg = children_class.dict_to_object(argument)

                # If it succeeds we have the right
                # class and instantiated object
                instantiated = True
            except KeyError:
                # This is not the right class, we should go see the parent
                classes.remove(children_class)
    elif origin in [list, collections.abc.Iterator]:
        # Homogenous sequences (lists)
        sequence_subtype = args[0]
        deserialized_arg = [deserialize_argument(sequence_subtype, arg) for arg in argument]
        if origin is collections.abc.Iterator:
            deserialized_arg = iter(deserialized_arg)

    elif origin is tuple:
        # Heterogenous sequences (tuples)
        deserialized_arg = tuple(deserialize_argument(t, arg) for (t, arg) in zip(args, argument))
    elif origin is dict:
        # Dynamic dict
        deserialized_arg = argument
    elif origin is InstanceOf:
        classname = args[0]
        object_class = dc.full_classname(object_=classname, compute_for='class')
        class_ = dcty.get_python_class_from_class_name(object_class)
        deserialized_arg = class_.dict_to_object(argument)
    else:
        msg = "Deserialization of typing {} is not implemented"
        raise NotImplementedError(msg.format(type_))
    return deserialized_arg


def deserialize_argument(type_, argument):
    """
    Deserialize an argument of a function with the type
    """
    if argument is None:
        return None
    if dcty.is_typing(type_):
        return deserialize_with_typing(type_, argument)

    if type_ in [TextIO, BinaryIO] or isinstance(argument, (StringFile, BinaryFile)):
        return argument

    if type_ in dcty.TYPING_EQUIVALENCES:
        if isinstance(argument, type_):
            return argument
        if isinstance(argument, int) and type_ == float:
            # Explicit conversion in this case
            return float(argument)
        # else ...
        msg = f"Given built-in type and argument are incompatible: " \
              f"{type(argument)} and {type_} in {argument}"
        raise TypeError(msg)

    if type_ is Any:
        # Any type
        return argument
    if inspect.isclass(type_) and issubclass(type_, (dc.DessiaObject, dessia_common.core.DessiaObject)):
        # Custom classes
        return type_.dict_to_object(argument)

    raise TypeError(f"Deserialization of ype {type_} is Not Implemented")


def find_references(value, path='#'):
    """
    Traverse recursively the value to find reference (pointers) in it
    Calls recursively find_references_sequence and find_references_dict
    """
    if isinstance(value, dict):
        return find_references_dict(value, path)
    if dcty.isinstance_base_types(value):
        return []
    if dcty.is_sequence(value):
        return find_references_sequence(value, path)
    if isinstance(value, (BinaryFile, StringFile)):
        return []
    raise ValueError(value)


def find_references_sequence(seq, path):
    if isinstance(seq, str):
        raise ValueError

    references = []
    for ie, element in enumerate(seq):
        path_value = f'{path}/{ie}'
        references.extend(find_references(element, path=path_value))
        # if value_nodes or value_edges:

    return references


def find_references_dict(dict_, path):
    if '$ref' in dict_:

        return [(path, dict_['$ref'])]

    references = []
    for key, value in dict_.items():
        if not dcty.isinstance_base_types(value):
            path_value = f'{path}/{key}'
            refs_value = find_references(value, path=path_value)
            references.extend(refs_value)
    return references


def pointer_graph(value):
    """
    Create a graph of subattributes of an object with edge representing either:
     * the hierarchy of an subattribute to an attribute
     * the pointer link between the 2 elements
    """

    nodes = set()
    path_edges = set()
    pointer_edges = set()

    # refs = []
    refs_by_path = {}

    for path, reference in find_references(value):
        # refs.append(reference)
        refs_by_path[path] = reference
        # Slitting path & reference to add missing nodes
        # For path
        path_segments = path.split('/')
        nodes.add(path_segments[0])
        previous_node = path_segments[0]
        for segment in path_segments[1:]:
            node = f'{previous_node}/{segment}'
            nodes.add(node)
            path_edges.add((previous_node, node))
            previous_node = node

        # For reference
        reference_segments = reference.split('/')
        nodes.add(reference_segments[0])
        previous_node = reference_segments[0]
        for segment in reference_segments[1:]:
            node = f'{previous_node}/{segment}'
            nodes.add(node)
            path_edges.add((previous_node, node))
            previous_node = node

        pointer_edges.add((path, reference))

    graph = nx.DiGraph()
    graph.add_nodes_from(nodes)
    graph.add_edges_from(path_edges, path=True)
    graph.add_edges_from(pointer_edges, head_type='circle', color='#008000', path=False)

    references = find_references(value)
    refs_by_path = {r[0]: r[1] for r in references}
    extra_edges = []
    new_paths = {}
    for path, point_at in references:  # loop over all refs?
        current_node = point_at
        last_name_change = None

        while current_node != '#':
            if current_node in refs_by_path:

                extra_edges.append((path, refs_by_path[current_node]))
                last_name_change = (current_node, refs_by_path[current_node])

            # Going up
            current_node = '/'.join(current_node.split('/')[:-1])

        if last_name_change:
            # marking as new path from top level pointer
            segments = point_at.split('/')
            for i in range(len(segments) - 1):
                node = '/'.join(segments[:i + 2])
                new_path = node.replace(*last_name_change)
                if new_path != node:
                    new_paths[node] = new_path
                    extra_edges.append((node, new_paths[node]))

    graph.add_edges_from(extra_edges, path=False)

    nx.set_node_attributes(graph, new_paths, "equals")

    return graph


def update_pointers_data(global_dict, current_dict, pointers_memo):
    if global_dict is None or pointers_memo is None:
        global_dict = current_dict

        if pointers_memo is None:
            pointers_memo = {}

        pointers_memo.update(dereference_jsonpointers(current_dict))
    return global_dict, pointers_memo


def deserialization_order(dict_):
    """
    Analyse a dict representing an object and give a deserialization order
    """
    graph = pointer_graph(dict_)
    if '#' not in graph.nodes:
        return []

    cycles = list(nx.simple_cycles(graph))
    if cycles:
        for cycle in cycles:
            print(cycle)
        raise NotImplementedError('Cycles in jsonpointers not handled')

    order = list(explore_tree_from_leaves(graph))
    if '#' in order:
        order.remove('#')

    cleaned_order = []
    equals = nx.get_node_attributes(graph, 'equals')

    for node in order:
        if node in equals:
            if not equals[node] in cleaned_order:
                cleaned_order.append((node, equals[node]))
        else:
            cleaned_order.append((node, None))
    return cleaned_order


def dereference_jsonpointers(dict_):  # , global_dict):
    """
    Analyse the given dict to:
    - find jsonpointers
    - deserialize them in the right order to respect pointers graph
    :returns: a dict with key the path of the item and the value is the python object
    """
    order = deserialization_order(dict_)
    pointers_memo = {}
    for node, reference in order:
        if reference is None:
            serialized_element = get_in_object_from_path(object_=dict_, path=node, evaluate_pointers=False)
            pointers_memo[node] = deserialize(serialized_element=serialized_element, global_dict=dict_,
                                              pointers_memo=pointers_memo, path=node)
        else:
            pointers_memo[node] = pointers_memo[reference]
    return pointers_memo


def pointer_graph_elements(value, path='#'):

    if isinstance(value, dict):
        return pointer_graph_elements_dict(value, path)
    if dcty.isinstance_base_types(value):
        return [], []
    if dcty.is_sequence(value):
        return pointer_graph_elements_sequence(value, path)

    raise ValueError(value)


def pointer_graph_elements_sequence(seq, path='#'):
    """
    Compute
    """
    if isinstance(seq, str):
        raise ValueError

    edges = []
    nodes = []
    for ie, element in enumerate(seq):
        path_value = f'{path}/{ie}'
        value_nodes, value_edges = pointer_graph_elements(element, path=path_value)
        # if value_nodes or value_edges:

        nodes.append(path_value)
        nodes.extend(value_nodes)

        edges.append((path, path_value))
        edges.extend(value_edges)

    return nodes, edges


def pointer_graph_elements_dict(dict_, path='#'):

    if '$ref' in dict_:
        return [path, dict_['$ref']], [(path, dict_['$ref'], True)]

    edges = []
    nodes = []
    for key, value in dict_.items():
        if not dcty.isinstance_base_types(value):
            path_value = f'{path}/{key}'
            value_nodes, value_edges = pointer_graph_elements(value, path=path_value)
            # if value_nodes or value_edges:
            nodes.append(path_value)
            nodes.extend(value_nodes)

            edges.append((path, path_value, False))
            edges.extend(value_edges)

    return nodes, edges


def pointers_analysis(obj):
    """
    Analyse on object to output stats on pointer use in the object
    :returns: a tuple of 2 dicts: one giving the number of pointer use by class
    """
    if isinstance(obj, dict):
        dict_ = obj
    else:
        dict_ = obj.to_dict()

    class_number = {}
    composed_by = {}
    class_from_path = {}
    graph = pointer_graph(dict_)
    for path1, path2 in graph.edges():
        if path1 != '#':
            if path2 in class_from_path:
                val2_class = class_from_path[path2]
            else:
                val2 = get_in_object_from_path(obj, path2)
                val2_class = dcty.full_classname(val2)
                class_from_path[path2] = val2_class

            class_number[val2_class] = class_number.get(val2_class, 0) + 1

            if path1 in class_from_path:
                val1_class = class_from_path[path1]
            else:
                val1 = get_in_object_from_path(obj, path1)
                val1_class = dcty.full_classname(val1)
                # val1_class = val1['object_class']
                class_from_path[path1] = val1_class

            if val1_class != val2_class:
                if val2_class not in composed_by:
                    composed_by[val2_class] = {}

                if val1_class not in composed_by[val2_class]:
                    composed_by[val2_class][val1_class] = 1
                else:
                    composed_by[val2_class][val1_class] += 1

    return class_number, composed_by



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu July 7 14:52:32 2022

@author: xavier-ferry
"""

from typing import List, Dict


class ToScriptElement:
    """
    Class meant to improve to_script readability
    """
    def __init__(self, declaration: str, before_declaration: str = None,
                 imports: List[str] = None, imports_as_is: List[str] = None):
        self.before_declaration = before_declaration
        self.declaration = declaration
        self.imports = imports
        self.imports_as_is = imports_as_is

    def imports_to_str(self) -> str:
        script_imports = ""
        for module, class_list in self.get_import_dict().items():
            script_imports += f"from {module} import {', '.join(class_list)}\n"

        for import_as_is in self.imports_as_is:
            import_str = f"import {import_as_is}\n"
            if import_str not in script_imports:
                script_imports += import_str

        return script_imports

    def get_import_dict(self) -> Dict[str, List[str]]:
        imports_dict: Dict[str, List[str]] = {}
        for import_ in self.imports:
            module = '.'.join(import_.split('.')[:-1])
            class_ = import_.split('.')[-1]
            if module not in imports_dict :
                imports_dict[module] = [class_]
            else:
                if class_ not in imports_dict[module]:
                    imports_dict[module].append(class_)

        return imports_dict



"""
generation for dessia_common

"""
from typing import List
import dectree as dt
import dessia_common.core as dc


class Generator(dc.DessiaObject):
    """
    Common parts of generator
    """

    def __init__(self, name: str = ''):
        dc.DessiaObject.__init__(self, name=name)

    def is_model_valid(self, model) -> bool:
        """
        Checks if Model is valid or not

        :param model: Model generated

        :return: True if valid and False othewise
        :rtype: bool
        """
        raise NotImplementedError('the method is_model_valid must be overloaded by subclassing class')

    def number_possibilities_from_model(self, model):
        raise NotImplementedError('the method number_possibilities_from_model must be overloaded by subclassing class')


class TreeGenerator(Generator):
    """
    Common parts of generator
    """
    _allowed_methods = ['generate']
    _non_serializable_attributes = ['tree']
    _non_data_eq_attributes = ['tree']

    def __init__(self, tree, name: str = ''):
        self.tree = tree
        Generator.__init__(self, name=name)

    def model_from_vector(self, vector: List[int]):
        """
        Generates the physical model from vector node
        :param vector: decision tree node vector
        :type vector: List[int]
        :return: model
        """
        raise NotImplementedError('the method model_from_vector must be overloaded by subclassing class')

    def is_vector_valid(self, vector: List[int]) -> bool:
        """
        Verifies and Validates the decision tree vector

        :param vector: decision tree node list
        :type vector: List[int]

        :return: True if valid and False otherwise
        :rtype: bool
        """
        return True

    def number_possibilities_from_vector(self, vector: List[int]):
        """
        This method is generic but can be overloaded to avoid model instanciation
        """
        model = self.model_from_vector(vector)
        return self.number_possibilities_from_model(model)


class DecisionTreeGenerator(TreeGenerator):
    """
    Abstract class, to be subclassed by real class
    This is still experimental and might be buggy
    """

    def __init__(self, name: str = ''):
        """
        :param name: The name of the generator
        """
        tree = dt.DecisionTree()
        TreeGenerator.__init__(self, tree, name=name)

    def generate(self, verbose: bool = False):
        model = self.model_from_vector(self.tree.current_node)

        self.tree.SetCurrentNodeNumberPossibilities(self.number_possibilities_from_model(model))
        while not self.tree.finished:
            if verbose:
                print('current node: ', self.tree.current_node)
            valid = self.is_vector_valid(self.tree.current_node)
            if valid:
                model = self.model_from_vector(self.tree.current_node)
                valid = self.is_model_valid(model)

            if verbose:
                print('node validity:', valid)
            if valid:
                number_possibilities = self.number_possibilities_from_model(model)
                if verbose:
                    print('number possibilities', number_possibilities)
                self.tree.SetCurrentNodeNumberPossibilities(number_possibilities)

            # TODO create a function in dectreee to know if a leaf
            if valid and (number_possibilities == 0):
                yield model

            self.tree.NextNode(valid)


class RegularDecisionTreeGenerator(TreeGenerator):
    """
    Abstract class, to be subclassed by real class
    This is still experimental and might be buggy
    """

    def __init__(self, number_possibilities: List[int], name: str = ''):
        self.number_possibilities = number_possibilities
        self.leaves_depth = len(self.number_possibilities) - 1
        tree = dt.RegularDecisionTree(number_possibilities)
        TreeGenerator.__init__(self, tree=tree, name=name)

    def generate(self, sorted_nodes: bool = False,
                 unique_nodes: bool = False,
                 verbose: bool = False):
        """

        Parameters
        ----------
        sorted_nodes : bool, optional
            DESCRIPTION. The default is False.
        unique_nodes : bool, optional
            DESCRIPTION. The default is False.
        verbose : bool, optional
            DESCRIPTION. The default is False.

        Yields
        ------
        model : TYPE
            DESCRIPTION.

        """
        if sorted_nodes:
            if unique_nodes:
                next_node_function = self.tree.NextSortedUniqueNode
            else:
                next_node_function = self.tree.NextSortedNode
        else:
            if unique_nodes:
                next_node_function = self.tree.NextUniqueNode
            else:
                next_node_function = self.tree.NextNode

        while not self.tree.finished:
            valid = self.is_vector_valid(self.tree.current_node)
            if verbose:
                print('current node: ', self.tree.current_node)
                print('node vector validity:', valid)
            if valid:
                model = self.model_from_vector(self.tree.current_node)
                valid = self.is_model_valid(model)
                if verbose:
                    print('node model validity:', valid)

                # TODO create a function in dectreee to know if a leaf
                if valid and self.tree.current_depth == self.leaves_depth:
                    yield model

            next_node_function(valid)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Types tools

"""
from typing import Any, Dict, List, Tuple, Type, Union, get_origin, get_args

import json
from collections.abc import Iterator, Sequence
from importlib import import_module

import dessia_common as dc
from dessia_common.typings import Subclass, InstanceOf, MethodType, ClassMethodType
from dessia_common.files import BinaryFile, StringFile


TYPING_EQUIVALENCES = {int: 'number', float: 'number', bool: 'boolean', str: 'string'}

TYPES_STRINGS = {int: 'int', float: 'float', bool: 'boolean', str: 'str',
                 list: 'list', tuple: 'tuple', dict: 'dict'}

SEQUENCE_TYPINGS = ['List', 'Sequence', 'Iterable']

TYPES_FROM_STRING = {'unicode': str, 'str': str, 'float': float, 'int': int, 'bool': bool}

SERIALIZED_BUILTINS = ['float', 'builtins.float', 'int', 'builtins.int', 'str', 'builtins.str', 'bool', 'builtins.bool']

_PYTHON_CLASS_CACHE = {}


def full_classname(object_, compute_for: str = 'instance'):
    if compute_for == 'instance':
        return object_.__class__.__module__ + '.' + object_.__class__.__name__
    if compute_for == 'class':
        return object_.__module__ + '.' + object_.__name__

    msg = 'Cannot compute {} full classname for object {}'
    raise NotImplementedError(msg.format(compute_for, object_))


def is_jsonable(x):
    """
    returns if object can be dumped as it is in a json
    """
    try:
        json.dumps(x)
        return True
    except TypeError:
        return False


def is_serializable(obj):
    if is_jsonable(obj):
        return True
    if isinstance(obj, dc.DessiaObject):
        dict_ = obj.to_dict()
        return is_jsonable(dict_)
    if isinstance(obj, dict):
        for key, value in obj.items():
            if not is_serializable(key) or not is_serializable(value):
                return False
        return True
    if is_sequence(obj):
        for element in obj:
            if not is_serializable(element):
                return False
        return True
    return False


def is_sequence(obj):
    """
    :param obj: Object to check
    :return: bool. True if object is a sequence but not a string.
                   False otherwise
    """
    return isinstance(obj, Sequence) and not isinstance(obj, str)


def is_builtin(type_):
    return type_ in TYPING_EQUIVALENCES


def isinstance_base_types(obj):
    """
    returns True if the object is either a str, a float a int or None
    """
    return isinstance(obj, (str, float, int)) or (obj is None)


def get_python_class_from_class_name(full_class_name):
    cached_value = _PYTHON_CLASS_CACHE.get(full_class_name, None)
    if cached_value is not None:
        return cached_value

    module_name, class_name = full_class_name.rsplit('.', 1)
    module = import_module(module_name)
    class_ = getattr(module, class_name)

    # Storing in cache
    _PYTHON_CLASS_CACHE[full_class_name] = class_
    return class_


def unfold_deep_annotation(typing_=None):
    if is_typing(typing_):
        origin = get_origin(typing_)
        args = get_args(typing_)
        return origin, args
    return None, None


def is_typing(object_: Any):
    has_module = hasattr(object_, '__module__')
    has_origin = hasattr(object_, '__origin__')
    has_args = hasattr(object_, '__args__')
    return has_module and has_origin and has_args


def serialize_typing(typing_):
    if is_typing(typing_):
        return serialize_typing_types(typing_)
    if typing_ in [StringFile, BinaryFile, MethodType, ClassMethodType] or isinstance(typing_, type):
        return full_classname(typing_, compute_for='class')
    return str(typing_)


def serialize_typing_types(typing_):
    origin = get_origin(typing_)
    args = get_args(typing_)
    if origin is Union:
        return serialize_union_typing(args)
    if origin is list:
        return f"List[{type_fullname(args[0])}]"
    if origin is tuple:
        argnames = ', '.join([type_fullname(a) for a in args])
        return f'Tuple[{argnames}]'
    # if origin is Iterator:
    #     return f"Iterator[{type_fullname(args[0])}]"
    if origin is dict:
        key_type = type_fullname(args[0])
        value_type = type_fullname(args[1])
        return f'Dict[{key_type}, {value_type}]'
    if origin is InstanceOf:
        return f'InstanceOf[{type_fullname(args[0])}]'
    if origin is Subclass:
        return f'Subclass[{type_fullname(args[0])}]'
    if origin is MethodType:
        return f'MethodType[{type_fullname(args[0])}]'
    if origin is ClassMethodType:
        return f'ClassMethodType[{type_fullname(args[0])}]'
    if origin is type:
        return "typing.Type"
    raise NotImplementedError(f"Serialization of typing {typing_} is not implemented")


def serialize_union_typing(args):
    if len(args) == 2 and type(None) in args:
        # This is a false Union => Is a default value set to None
        return serialize_typing(args[0])

    # Types union
    argnames = ', '.join([type_fullname(a) for a in args])
    return f'Union[{argnames}]'


def type_fullname(arg):
    if arg.__module__ == 'builtins':
        full_argname = '__builtins__.' + arg.__name__
    else:
        full_argname = serialize_typing(arg)
    return full_argname


def type_from_argname(argname):
    splitted_argname = argname.rsplit('.', 1)
    if argname:
        if splitted_argname[0] == '__builtins__':
            argname = f"builtins.{splitted_argname[1]}"
        # if splitted_argname[0] != '__builtins__':
        return get_python_class_from_class_name(argname)
        # return literal_eval(splitted_argname[1])
    return Any


TYPING_FROM_SERIALIZED_NAME = {"List": List, "Tuple": Tuple, "Iterator": Iterator, "Dict": Dict}


def deserialize_typing(serialized_typing):
    # TODO : handling recursive deserialization
    if isinstance(serialized_typing, str):
        # TODO other builtins should be implemented
        if serialized_typing in SERIALIZED_BUILTINS:
            return deserialize_builtin_typing(serialized_typing)

        if serialized_typing in ["dessia_common.files.StringFile", "dessia_common.files.BinaryFile"]:
            return deserialize_file_typing(serialized_typing)

        if serialized_typing in ["dessia_common.typings.MethodType", "dessia_common.typings.ClassMethodType"]:
            return deserialize_method_typing(serialized_typing)

        if serialized_typing == "Type":
            return Type

        if '[' in serialized_typing:
            toptype, remains = serialized_typing.split('[', 1)
            full_argname = remains.rsplit(']', 1)[0]
            if "[" in full_argname:
                return TYPING_FROM_SERIALIZED_NAME[toptype][deserialize_typing(full_argname)]
        else:
            toptype = serialized_typing
            full_argname = ''
        if toptype == 'List':
            return List[type_from_argname(full_argname)]
        if toptype == 'Tuple':
            return deserialize_tuple_typing(full_argname)
        # if toptype == "Iterator":
        #     return Iterator[type_from_argname(full_argname)]
        if toptype == 'Dict':
            args = full_argname.split(', ')
            key_type = type_from_argname(args[0])
            value_type = type_from_argname(args[1])
            return Dict[key_type, value_type]
        if toptype == "InstanceOf":
            return InstanceOf[type_from_argname(full_argname)]
        # if toptype == "Subclass":
        #     return InstanceOf[type_from_argname(full_argname)]
        return get_python_class_from_class_name(serialized_typing)
    raise NotImplementedError(f'{serialized_typing} of type {type(serialized_typing)}')


def deserialize_tuple_typing(full_argname):
    if ', ' in full_argname:
        args = full_argname.split(', ')
        if len(args) == 0:
            return Tuple
        if len(args) == 1:
            type_ = type_from_argname(args[0])
            return Tuple[type_]
        if len(set(args)) == 1:
            type_ = type_from_argname(args[0])
            return Tuple[type_, ...]

        raise TypeError("Heterogenous tuples are forbidden as types for workflow non-block variables.")
    return Tuple[type_from_argname(full_argname)]


def deserialize_file_typing(serialized_typing):
    if serialized_typing == "dessia_common.files.StringFile":
        return StringFile
    if serialized_typing == "dessia_common.files.BinaryFile":
        return BinaryFile
    raise NotImplementedError(f"File typing {serialized_typing} deserialization is not implemented")


def deserialize_method_typing(serialized_typing):
    if serialized_typing == "dessia_common.typings.MethodType":
        return MethodType
    if serialized_typing == "dessia_common.typings.ClassMethodType":
        return ClassMethodType
    raise NotImplementedError(f"Method typing {serialized_typing} deserialization is not implemented")


def deserialize_builtin_typing(serialized_typing):
    if serialized_typing in ['float', 'builtins.float']:
        return float
    if serialized_typing in ['int', 'builtins.int']:
        return int
    if serialized_typing in ['str', 'builtins.str']:
        return str
    if serialized_typing in ['bool', 'builtins.bool']:
        return bool
    raise NotImplementedError(f"Builtin typing of {serialized_typing} deserialization is not implemented")


def is_bson_valid(value, allow_nonstring_keys=False) -> Tuple[bool, str]:
    """
    returns validity (bool) and a hint (str)
    """
    if isinstance(value, (int, float, str)):
        return True, ''

    if value is None:
        return True, ''

    if isinstance(value, dict):
        for k, subvalue in value.items():
            # Key check
            if isinstance(k, str):
                if '.' in k:
                    return False, f'key {k} of dict is a string containing a ., which is forbidden'
            elif isinstance(k, float):
                return False, f'key {k} of dict is a float, which is forbidden'
            elif isinstance(k, int):
                if not allow_nonstring_keys:
                    log = f'key {k} of dict is an unsuported type {type(k)},' \
                          f'use allow_nonstring_keys=True to allow'
                    return False, log
            else:
                return False, f'key {k} of dict is an unsuported type {type(k)}'

            # Value Check
            v_valid, hint = is_bson_valid(value=subvalue, allow_nonstring_keys=allow_nonstring_keys)
            if not v_valid:
                return False, hint

    elif is_sequence(value):
        for subvalue in value:
            valid, hint = is_bson_valid(value=subvalue, allow_nonstring_keys=allow_nonstring_keys)
            if not valid:
                return valid, hint
    else:
        return False, f'Unrecognized type: {type(value)}'
    return True, ''

# TODO recursive_type and recursive_type functions look weird


def recursive_type(obj):
    """
    What is the difference with serialize typing?
    """

    if isinstance(obj, tuple(list(TYPING_EQUIVALENCES.keys()) + [dict])):
        type_ = TYPES_STRINGS[type(obj)]
    elif isinstance(obj, dc.DessiaObject):
        type_ = obj.__module__ + '.' + obj.__class__.__name__
    elif hasattr(obj, 'output_type'):
        type_ = obj.output_type
    elif isinstance(obj, (list, tuple)):
        type_ = []
        for element in obj:
            type_.append(recursive_type(element))
    elif obj is None:
        type_ = None
    else:
        raise NotImplementedError(obj)
    return type_


def union_is_default_value(typing_: Type) -> bool:
    """
    Union typings can be False positives.
    An argument of a function that has a default_value set to None is Optional[T],
    which is an alias for Union[T, NoneType]. This function checks if this is the case.
    """
    args = get_args(typing_)
    return len(args) == 2 and type(None) in args


def typematch(type_: Type, match_against: Type) -> bool:
    """
        Return wether type_ matches against match_against.
        match_against needs to be "wider" than type_, and the check is not bilateral
    """
    # TODO Implement a more intelligent check for Unions : Union[T, U] should match against Union[T, U, V]
    # TODO Implement a check for Dict
    if type_ == match_against or match_against is Any or particular_typematches(type_, match_against):
        # Trivial cases. If types are strictly equal, then it should pass straight away
        return True

    if is_typing(type_):
        return complex_first_type_match(type_, match_against)

    if not is_typing(match_against):
        # type_ and match_against aren't complex : check for subclass only
        if issubclass(type_, match_against):
            return True

    # type_ is not complex and match_against is
    match_against, origin, args = heal_type(match_against)
    if origin is Union:
        matches = [typematch(type_, subtype) for subtype in args]
        return any(matches)
    return False


def complex_first_type_match(type_: Type, match_against: Type) -> bool:
    """
    Match type when type_ is a complex typing (List, Union, Tuple,...)
    """
    # Complex typing for the first type_. Cases : List, Tuple, Union
    if not is_typing(match_against):
        # Type matching is unilateral and match against should be more open than type_
        return False

    # Inspecting and healing types
    type_, type_origin, type_args = heal_type(type_)
    match_against, match_against_origin, match_against_args = heal_type(match_against)

    if type_origin != match_against_origin:
        # Being strict for now. Is there any other case than default values where this would be wrong ?
        return False

    if type_origin is list:
        # Can only have one arg, should match
        return typematch(type_args[0], match_against_args[0])

    if type_origin is tuple:
        # Order matters, all args should match
        return all(typematch(a, b) for a, b in zip(type_args, match_against_args))

    if type_origin is dict:
        # key type AND value type should match
        return typematch(type_args[0], match_against_args[0]) and typematch(type_args[1], match_against_args[1])

    if type_origin is Union:
        # type args must be a subset of match_against args set
        type_argsset = set(type_args)
        match_against_argsset = set(match_against_args)
        return type_argsset.issubset(match_against_argsset)

    # Otherwise, it is not implemented
    raise NotImplementedError(f"Type {type_} is a complex typing and cannot be matched against others yet")


def heal_type(type_: Type):
    """
    Inspect type and returns its params

    For now, only checks wether the type is an 'Optional' / Union[T, NoneType],
    which should be flattened and not considered

    returns the cleaned type, origin and args
    """
    type_origin = get_origin(type_)
    type_args = get_args(type_)
    if type_origin is Union:
        # Check for default values false positive
        if union_is_default_value(type_):
            type_ = type_args[0]
            type_origin = get_origin(type_)
            type_args = get_args(type_)
    return type_, type_origin, type_args


def particular_typematches(type_: Type, match_against: Type) -> bool:
    """
    Checks for specific cases of typematches and returns and boolean
    """
    if type_ is int and match_against is float:
        return True
    return False



from dessia_common.vectored_objects import Catalog, ParetoSettings
import dessia_common.workflow as wf

choice_args = ['MPG', 'Cylinders', 'Displacement', 'Horsepower',
               'Weight', 'Acceleration', 'Model']  # Ordered

minimized_attributes = {'MPG': False, 'Horsepower': True,
                        'Weight': True, 'Acceleration': False}

aimpoint = {'MPG': 4, 'Weight': 1000}

filter_min = [{'attribute': 'MPG', 'comparison_operator': 'lte', 'bound': 20}]

# Blocks
import_csv = wf.Import(type_='csv')
instantiate_pareto = wf.InstanciateModel(model_class=ParetoSettings,
                                         name='Pareto Settings')
catalog1 = wf.InstanciateModel(model_class=Catalog,
                               name='Cars 1')
filter1 = wf.ModelMethod(model_class=Catalog,
                         method_name='filter_',
                         name='Filter 1')
filtered_catalog1 = wf.InstanciateModel(model_class=Catalog,
                                        name='Filtered Catalog 1')

filter_max = [{'attribute': 'MPG', 'comparison_operator': 'gte', 'bound': 30}]
catalog2 = wf.InstanciateModel(model_class=Catalog,
                               name='Car 2')
filter2 = wf.ModelMethod(model_class=Catalog,
                         method_name='filter_',
                         name='Filter 2')
filtered_catalog2 = wf.InstanciateModel(model_class=Catalog,
                                        name='Filtered Catalog 2')

catalogs = wf.Sequence(number_arguments=2, name='Catalogs')
concatenate = wf.ClassMethod(class_=Catalog, method_name='concatenate',
                             name='Concatenate')
display = wf.Display(order=0)

array_attribute = wf.ModelAttribute(attribute_name='array', name='Array')

blocks = [import_csv, instantiate_pareto,
          catalog1, filter1, filtered_catalog1,
          catalog2, filter2, filtered_catalog2,
          catalogs, concatenate, display, array_attribute]

# Pipes
pipes = [
    wf.Pipe(import_csv.outputs[0], catalog1.inputs[0]),
    wf.Pipe(import_csv.outputs[1], catalog1.inputs[1]),
    wf.Pipe(instantiate_pareto.outputs[0], catalog1.inputs[2]),
    wf.Pipe(import_csv.outputs[0], catalog2.inputs[0]),
    wf.Pipe(import_csv.outputs[1], catalog2.inputs[1]),
    wf.Pipe(instantiate_pareto.outputs[0], catalog2.inputs[2]),
    wf.Pipe(catalog1.outputs[0], filter1.inputs[0]),
    wf.Pipe(catalog2.outputs[0], filter2.inputs[0]),
    wf.Pipe(filter1.outputs[0], filtered_catalog1.inputs[0]),
    wf.Pipe(import_csv.outputs[1], filtered_catalog1.inputs[1]),
    wf.Pipe(instantiate_pareto.outputs[0], filtered_catalog1.inputs[2]),
    wf.Pipe(filtered_catalog1.outputs[0], catalogs.inputs[0]),
    wf.Pipe(filter2.outputs[0], filtered_catalog2.inputs[0]),
    wf.Pipe(import_csv.outputs[1], filtered_catalog2.inputs[1]),
    wf.Pipe(instantiate_pareto.outputs[0], filtered_catalog2.inputs[2]),
    wf.Pipe(filtered_catalog2.outputs[0], catalogs.inputs[1]),
    wf.Pipe(catalogs.outputs[0], concatenate.inputs[0]),
    wf.Pipe(import_csv.outputs[1], concatenate.inputs[3]),
    wf.Pipe(concatenate.outputs[0], display.inputs[0]),
    wf.Pipe(concatenate.outputs[0], array_attribute.inputs[0])
]

# Workflow
workflow = wf.Workflow(blocks=blocks, pipes=pipes,
                       output=catalogs.outputs[0],
                       name='Cars workflow')
# workflow.plot_jointjs()

# # Input values
input_values = {
    workflow.index(import_csv.inputs[0]): 'cars.csv',
    workflow.index(instantiate_pareto.inputs[0]): minimized_attributes,
    workflow.index(instantiate_pareto.inputs[1]): True,
    workflow.index(catalog1.inputs[4]): choice_args,
    workflow.index(catalog1.inputs[5]): 'Cars 1',
    workflow.index(filter1.inputs[1]): filter_min,
    workflow.index(filtered_catalog1.inputs[4]): choice_args,
    workflow.index(filtered_catalog1.inputs[5]): 'Filtered Cars 1',
    workflow.index(catalog2.inputs[4]): choice_args,
    workflow.index(catalog2.inputs[5]): 'Cars 2',
    workflow.index(filter2.inputs[1]): filter_max,
    workflow.index(filtered_catalog2.inputs[4]): choice_args,
    workflow.index(filtered_catalog2.inputs[5]): 'Filtered Cars 2',
}
workflow_run = workflow.run(input_values=input_values)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Models for RandData objects data
"""

from dessia_common.tests import RandDataD1, RandDataD2, RandDataD3, RandDataD4, RandDataD5, RandDataD6, RandDataD7,\
    RandDataD8, RandDataD9, RandDataD10

# Tests RandDatas
mean_borns = (-50, 50)
std_borns = (-2, 2)

rand_data_d1 = RandDataD1.create_dataset(nb_clusters=10, nb_points=250, mean_borns=mean_borns, std_borns=std_borns)
rand_data_d2 = RandDataD2.create_dataset(nb_clusters=10, nb_points=250, mean_borns=mean_borns, std_borns=std_borns)
rand_data_d3 = RandDataD3.create_dataset(nb_clusters=10, nb_points=250, mean_borns=mean_borns, std_borns=std_borns)
rand_data_d4 = RandDataD4.create_dataset(nb_clusters=10, nb_points=250, mean_borns=mean_borns, std_borns=std_borns)
rand_data_d5 = RandDataD5.create_dataset(nb_clusters=10, nb_points=250, mean_borns=mean_borns, std_borns=std_borns)
rand_data_d6 = RandDataD6.create_dataset(nb_clusters=10, nb_points=250, mean_borns=mean_borns, std_borns=std_borns)
rand_data_d7 = RandDataD7.create_dataset(nb_clusters=10, nb_points=500, mean_borns=mean_borns, std_borns=std_borns)
rand_data_d8 = RandDataD8.create_dataset(nb_clusters=10, nb_points=500, mean_borns=mean_borns, std_borns=std_borns)
rand_data_d9 = RandDataD9.create_dataset(nb_clusters=10, nb_points=500, mean_borns=mean_borns, std_borns=std_borns)
rand_data_d10 = RandDataD10.create_dataset(nb_clusters=10, nb_points=250, mean_borns=mean_borns, std_borns=std_borns)

rand_data_small = rand_data_d5 + rand_data_d4 + rand_data_d3
rand_data_middl = rand_data_d6 + rand_data_d4 + rand_data_d5
rand_data_large = rand_data_d9 + rand_data_d7 + rand_data_d8



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
exports for dessia_common

"""
from openpyxl.styles.borders import Border, Side
from openpyxl.styles import PatternFill, Font
from openpyxl import Workbook
import openpyxl.utils
from dessia_common.breakdown import breakdown


def is_hashable(value):
    """Determine whether `value` can be hashed."""
    try:
        hash(value)
    except TypeError:
        return False
    return True


def is_number(value):
    """
    Determine if the value is a int or a float
    """
    return isinstance(value, (int, float))


def is_builtins_list(list_):
    """
    Determin if a list is only composed of builtins
    """
    for element in list_:
        if not (is_number(element) or isinstance(element, str)):
            return False
    return True


class XLSXWriter:

    max_column_width = 40
    color_dessIA1 = "263238"
    color_dessIA2 = "537CB0"
    grey1 = "f1f1f1"
    white_font = Font(color="FFFFFF")

    thin_border = Border(left=Side(style='thin'), right=Side(style='thin'),
                         top=Side(style='thin'), bottom=Side(style='thin'))

    def __init__(self, object_):
        """
        :param object_: an Dessiaobject to write as excel
        """

        self.pattern_color1 = PatternFill(
            fill_type="solid",
            start_color=self.color_dessIA1,
            end_color=self.color_dessIA1)

        self.pattern_color2 = PatternFill(
            fill_type="solid",
            start_color=self.color_dessIA2,
            end_color=self.color_dessIA2)

        self.workbook = Workbook()
        self.main_sheet = self.workbook.active
        self.object = object_

        self.paths = breakdown(object_)

        self.classes_to_sheets = {}
        self.object_to_sheet_row = {}
        for class_name, obj_paths in self.paths.items():
            sheet = self.workbook.create_sheet(class_name)
            self.classes_to_sheets[class_name] = sheet

            for i, (obj, path) in enumerate(obj_paths.items()):
                self.object_to_sheet_row[obj] = (sheet, i + 2, path)

        self.write()

    def write_class_header_to_row(self, obj_of_class, sheet, row_number):
        """
        Writes to a sheet the class header: finds columns names from a class
        """
        cell = sheet.cell(row=row_number, column=1, value='Path')
        cell.fill = self.pattern_color2
        cell.border = self.thin_border
        cell.font = self.white_font

        cell = sheet.cell(row=row_number, column=2, value='name')
        cell.fill = self.pattern_color2
        cell.border = self.thin_border
        cell.font = self.white_font
        i = 3

        for (k, _) in sorted(obj_of_class.__dict__.items()):
            if (not k.startswith('_')) and k != 'name':
                cell = sheet.cell(row=row_number, column=i, value=str(k))

                cell.border = self.thin_border
                cell.fill = self.pattern_color2
                cell.font = self.white_font
                i += 1

    def write_value_to_cell(self, value, sheet, row_number, column_number):
        """
        Write a given value to a cell. Insert it as a link if it is an object
        """
        cell_link = None
        if isinstance(value, dict):
            str_v = f'Dict of {len(value)} items'
        elif isinstance(value, list):
            if is_builtins_list(value):
                str_v = str(value)
            else:
                str_v = f'List of {len(value)} items'

        elif isinstance(value, set):
            str_v = f'Set of {len(value)} items'
        elif isinstance(value, float):
            str_v = round(value, 6)
        elif is_hashable(value) and value in self.object_to_sheet_row:
            ref_sheet, ref_row_number, ref_path = self.object_to_sheet_row[value]
            str_v = ref_path
            cell_link = f'#{ref_sheet.title}!A{ref_row_number}'
        else:
            str_v = str(value)

        cell = sheet.cell(row=row_number, column=column_number, value=str_v)
        if cell_link:
            cell.hyperlink = cell_link

        cell.border = self.thin_border

    def write_object_to_row(self, obj, sheet, row_number, path=''):
        """
        Write on object to a row. Loops on its attributes to write its value in each cell
        """
        cell = sheet.cell(row=row_number, column=1, value=path)
        cell.border = self.thin_border
        if hasattr(obj, 'name'):
            cell = sheet.cell(row=row_number, column=2, value=obj.name)
        else:
            cell = sheet.cell(row=row_number, column=2, value='No name in model')

        cell.border = self.thin_border
        i = 3
        for (k, value) in sorted(obj.__dict__.items()):
            if (not k.startswith('_')) and k != 'name':
                self.write_value_to_cell(value, sheet, row_number, i)

                i += 1

                # column_width = min((len(k) + 1.5), max_column_width)
                # column_name = openpyxl.utils.cell.get_column_letter(i)
                # sheet.column_dimensions[column_name].width = column_width

    def write_object_id(self, sheet):
        """
        Write object id to a given sheet
        """
        sheet.title = f'Object {self.object.__class__.__name__}'

        sheet['A1'] = 'Module'
        sheet['B1'] = 'Class'
        sheet['C1'] = 'name'

        sheet['A1'].border = self.thin_border
        sheet['B1'].border = self.thin_border
        sheet['C1'].border = self.thin_border
        sheet['A1'].fill = self.pattern_color1
        sheet['B1'].fill = self.pattern_color1
        sheet['C1'].fill = self.pattern_color1
        sheet['A1'].font = self.white_font
        sheet['B1'].font = self.white_font
        sheet['C1'].font = self.white_font

        sheet['A2'] = self.object.__module__
        sheet['B2'] = self.object.__class__.__name__
        sheet['C2'] = self.object.name

        sheet['A2'].border = self.thin_border
        sheet['B2'].border = self.thin_border
        sheet['C2'].border = self.thin_border
        sheet['A2'].fill = self.pattern_color1
        sheet['B2'].fill = self.pattern_color1
        sheet['C2'].fill = self.pattern_color1
        sheet['A2'].font = self.white_font
        sheet['B2'].font = self.white_font
        sheet['C2'].font = self.white_font

        sheet['A3'] = 'Attribute'
        sheet['A4'] = 'Value'
        sheet['A3'].border = self.thin_border
        sheet['A4'].border = self.thin_border

    def write(self):
        """
        Generate the whole file
        """
        # name_column_width = 0
        self.write_object_id(self.main_sheet)
        self.write_class_header_to_row(self.object, self.main_sheet, 3)
        self.write_object_to_row(self.object, self.main_sheet, 4)
        self.autosize_sheet_columns(self.main_sheet, 5, 30)

        for class_name, obj_paths in self.paths.items():
            sheet = self.classes_to_sheets[class_name]

            for obj, path in obj_paths.items():
                _, row_number, path = self.object_to_sheet_row[obj]
                self.write_object_to_row(obj, sheet, row_number, path)
            self.write_class_header_to_row(obj, sheet, 1)

            sheet.auto_filter.ref = f"A1:{openpyxl.utils.cell.get_column_letter(sheet.max_column)}{len(obj_paths) + 1}"
            self.autosize_sheet_columns(sheet, 5, 30)

    def save_to_file(self, filepath: str):
        """
        Save to a filepath (open) and write
        """
        if not filepath.endswith('.xlsx'):
            filepath += '.xlsx'
            print(f"Changing name to {filepath}")

        with open(filepath, 'rb') as file:
            self.save_to_stream(file)

    def save_to_stream(self, stream):
        """
        Saves the file to a binary stream
        """
        self.workbook.save(stream)

    @staticmethod
    def autosize_sheet_columns(sheet, min_width=5, max_width=30):
        """
        Autosize the sheet columns by analyzing the content. Min and max width must be specified
        """
        # Autosize columns
        for col in sheet.columns:
            width = min_width
            column = col[1].column_letter  # Get the column name
            # Since Openpyxl 2.6, the column name is  ".column_letter" as .column became the column number (1-based)
            for cell in col:
                try:  # Necessary to avoid error on empty cells
                    if len(str(cell.value)) > width:
                        width = len(str(cell.value))
                except AttributeError:
                    pass
            if width > 0:
                adjusted_width = min((width + 0.5), max_width)
                sheet.column_dimensions[column].width = adjusted_width



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov 24 19:24:53 2021

@author: steven
"""

import math
import numpy as npy
import dessia_common as dc
from dessia_common.utils.types import isinstance_base_types, is_sequence, full_classname
from dessia_common.files import BinaryFile, StringFile

def diff(value1, value2, path='#'):
    """
    Main function to get the diff of two objects
    :return: a tuple of a list of diff between the objects, missing keys in other object and invalid types
    """
    diff_values = []
    missing_keys_in_other_object = []
    invalid_types = []

    if is_sequence(value1) and is_sequence(value2):
        return sequence_diff(value1, value2, path=path)

    if not isinstance(value1, type(value2)):
        invalid_types.append(path)
        return diff_values, missing_keys_in_other_object, invalid_types

    if isinstance_base_types(value1):
        if isinstance(value1, float) and \
                math.isclose(value1, value2, abs_tol=dc.FLOAT_TOLERANCE):
            return diff_values, missing_keys_in_other_object, invalid_types
        if value1 != value2:
            diff_values.append((path, value1, value2))
        return diff_values, missing_keys_in_other_object, invalid_types
    if isinstance(value1, dict):
        return dict_diff(value1, value2, path=path)
    # elif hasattr(value1, '_data_eq'):

    # Should be object
    if hasattr(value1, '_data_eq'):
        # DessiaObject
        if value1._data_eq(value2):
            return [], [], []

        # Use same code snippet as in data_eq
        eq_dict = value1._serializable_dict()
        if 'name' in eq_dict:
            del eq_dict['name']

        other_eq_dict = value2._serializable_dict()
        return dict_diff(eq_dict, other_eq_dict, path=path)

    if value1 == value2:
        return [], [], []

    raise NotImplementedError(f'Undefined type in diff: {type(value1)}')


def dict_diff(dict1, dict2, path='#'):
    missing_keys_in_other_object = []
    diff_values = []
    invalid_types = []

    for key, value in dict1.items():
        path_key = f'{path}/{key}'
        if key not in dict2:
            missing_keys_in_other_object.append(key)
        else:
            diff_key, mkk, itk = diff(value, dict2[key], path=path_key)
            diff_values.extend(diff_key)
            missing_keys_in_other_object.extend(mkk)
            invalid_types.extend(itk)

    return diff_values, missing_keys_in_other_object, invalid_types


def sequence_diff(seq1, seq2, path='#'):
    diff_values = []
    missing_keys_in_other_object = []
    invalid_types = []

    if len(seq1) != len(seq2):
        diff_values.append((path, seq1, seq2))
    else:
        for i, (v1, v2) in enumerate(zip(seq1, seq2)):
            path_value = f'{path}/{i}'
            diff_value, missing_key_value, invalid_type_value = diff(v1, v2, path=path_value)
            # print('dvs', dv, v1, v2)
            diff_values.extend(diff_value)
            missing_keys_in_other_object.extend(missing_key_value)
            invalid_types.extend(invalid_type_value)
    return diff_values, missing_keys_in_other_object, invalid_types


def data_eq(value1, value2):
    if is_sequence(value1) and is_sequence(value2):
        return sequence_data_eq(value1, value2)

    if isinstance(value1, npy.int64) or isinstance(value2, npy.int64):
        return value1 == value2

    if isinstance(value1, npy.float64) or isinstance(value2, npy.float64):
        return math.isclose(value1, value2, abs_tol=dc.FLOAT_TOLERANCE)

    if not isinstance(value2, type(value1))\
            and not isinstance(value1, type(value2)):
        return False

    if isinstance_base_types(value1):
        if isinstance(value1, float):
            return math.isclose(value1, value2, abs_tol=dc.FLOAT_TOLERANCE)

        return value1 == value2

    if isinstance(value1, dict):
        return dict_data_eq(value1, value2)

    if isinstance(value1, (BinaryFile, StringFile)):
        return value1 == value2

    # Else: its an object

    if full_classname(value1) != full_classname(value2):
        # print('full classname !=')
        return False

    # Test if _data_eq is customized
    if hasattr(value1, '_data_eq'):
        custom_method = (value1._data_eq.__code__
                         is not dc.DessiaObject._data_eq.__code__)
        if custom_method:
            return value1._data_eq(value2)

    # Not custom, use generic implementation
    eq_dict = value1._serializable_dict()
    if 'name' in eq_dict:
        del eq_dict['name']

    other_eq_dict = value2._serializable_dict()

    return dict_data_eq(eq_dict, other_eq_dict)


def dict_data_eq(dict1, dict2):

    for key, value in dict1.items():
        if key not in dict2:
            return False
        if not data_eq(value, dict2[key]):
            return False
    return True


def sequence_data_eq(seq1, seq2):
    if len(seq1) != len(seq2):
        return False

    for v1, v2 in zip(seq1, seq2):
        if not data_eq(v1, v2):
            # print('seq false')
            return False

    return True


def choose_hash(object_):
    if is_sequence(object_):
        return list_hash(object_)
    if isinstance(object_, dict):
        return dict_hash(object_)
    if isinstance(object_, str):
        return sum(ord(e) for e in object_)
    return hash(object_)


def list_hash(list_):
    return sum(choose_hash(e) for e in list_)


def dict_hash(dict_):
    hash_ = 0
    for key, value in dict_.items():
        hash_ += hash(key) + choose_hash(value)
    return hash_



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Algebra functions
"""
from functools import reduce


def number2factor(number):
    """
    Temporary function : Add to some tools package
    Finds all the ways to combine elements
    """
    factor_range = range(1, int(number ** 0.5) + 1)

    if number:
        factors = list(set(reduce(list.__add__, ([i, number // i]
                                                 for i in factor_range
                                                 if number % i == 0))))

        grids = [(factor_x, int(number / factor_x))
                 for factor_x in factors
                 if (number / factor_x).is_integer()]
    else:
        grids = []
    return grids


def number3factor(number, complete=True):
    """
    Temporary function : Add to some tools package
    Finds all the ways to combine elements
    """
    factor_range = range(1, int(number ** 0.5) + 1)

    if number:
        factors = list(set(reduce(list.__add__, ([i, number // i]
                                                 for i in factor_range
                                                 if number % i == 0))))
        if not complete:
            grids = get_incomplete_factors(number, factors)

        else:
            grids = [(factor_x, factor_y, int(number / (factor_x * factor_y)))
                     for factor_x in factors
                     for factor_y in factors
                     if (number / (factor_x * factor_y)).is_integer()]
        return grids
    return []


def get_incomplete_factors(number, factors):
    """
    TODO
    """
    grids = []
    sets = []
    for factor_x in factors:
        for factor_y in factors:
            value = number / (factor_x * factor_y)
            if value.is_integer():
                grid = (factor_x, factor_y, int(value))
                if set(grid) not in sets:
                    sets.append(set(grid))
                    grids.append(grid)
    return grids



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 18 18:52:32 2022

@author: masfaraud
"""

import inspect
import warnings

from zipfile import ZipFile
from typing import List, Type, Any, Dict, get_type_hints

import itertools
from dessia_common.core import DessiaFilter, FiltersList, enhanced_deep_attr, split_argspecs,\
    type_from_annotation, DessiaObject
from dessia_common.utils.types import get_python_class_from_class_name, full_classname
from dessia_common.utils.docstrings import parse_docstring, EMPTY_PARSED_ATTRIBUTE
from dessia_common.utils.serialization import serialize
from dessia_common.errors import UntypedArgumentError
from dessia_common.typings import JsonSerializable, MethodType, ClassMethodType
from dessia_common.files import StringFile, BinaryFile
from dessia_common.utils.helpers import concatenate
from dessia_common.breakdown import attrmethod_getter

from dessia_common.workflow.core import Block, Variable, TypedVariable, TypedVariableWithDefaultValue,\
    set_block_variable_names_from_dict, Workflow, DisplaySetting, DisplayObject
from dessia_common.workflow.utils import ToScriptElement


def set_inputs_from_function(method, inputs=None):
    """
    Inspect given method argspecs and sets block inputs from it
    """
    if inputs is None:
        inputs = []
    args_specs = inspect.getfullargspec(method)
    nargs, ndefault_args = split_argspecs(args_specs)

    for iarg, argument in enumerate(args_specs.args):
        if argument not in ['self', 'cls', 'progress_callback']:
            try:
                annotations = get_type_hints(method)
                type_ = type_from_annotation(annotations[argument], module=method.__module__)
            except KeyError as error:
                raise UntypedArgumentError(f"Argument {argument} of method/function {method.__name__} has no typing")\
                    from error
            if iarg > nargs - ndefault_args:
                default = args_specs.defaults[ndefault_args - nargs + iarg - 1]
                input_ = TypedVariableWithDefaultValue(type_=type_, default_value=default, name=argument)
                inputs.append(input_)
            else:
                inputs.append(TypedVariable(type_=type_, name=argument))
    return inputs


def output_from_function(function, name: str = "result output"):
    """
    Inspects given function argspecs and compute block output from it
    """
    annotations = get_type_hints(function)
    if 'return' in annotations:
        type_ = type_from_annotation(annotations['return'], function.__module__)
        return TypedVariable(type_=type_, name=name)
    return Variable(name=name)


class BlockError(Exception):
    pass


class InstantiateModel(Block):
    """
    :param model_class: The class to instanciate.
    :type model_class: Instanciable
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, model_class: Type, name: str = ''):
        self.model_class = model_class
        inputs = []
        inputs = set_inputs_from_function(self.model_class.__init__, inputs)
        outputs = [TypedVariable(type_=self.model_class, name='Instanciated object')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return len(self.model_class.__name__)

    def equivalent(self, other):
        classname = self.model_class.__class__.__name__
        other_classname = other.model_class.__class__.__name__
        return Block.equivalent(self, other) and classname == other_classname

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_['model_class'] = full_classname(object_=self.model_class, compute_for='class')
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        if 'model_class_module' in dict_:  # TODO Retro-compatibility. Remove this in future versions
            module_name = dict_['model_class_module']
            classname = module_name + '.' + dict_['model_class']
        else:
            classname = dict_['model_class']
        class_ = get_python_class_from_class_name(classname)
        return cls(class_, name=dict_['name'])

    def evaluate(self, values):
        args = {var.name: values[var] for var in self.inputs}
        return [self.model_class(**args)]

    def package_mix(self):
        return {self.model_class.__module__.split('.')[0]: 1}

    def _docstring(self):
        docstring = self.model_class.__doc__
        annotations = get_type_hints(self.model_class.__init__)
        parsed_docstring = parse_docstring(docstring=docstring, annotations=annotations)
        parsed_attributes = parsed_docstring["attributes"]
        block_docstring = {i: parsed_attributes[i.name] if i.name in parsed_attributes
                           else EMPTY_PARSED_ATTRIBUTE for i in self.inputs}
        return block_docstring

    def _to_script(self) -> ToScriptElement:
        script = f"InstantiateModel(model_class=" \
                 f"{self.model_class.__name__}, name='{self.name}')"

        imports = [
            full_classname(object_=self.model_class, compute_for='class'),
            self.full_classname
        ]
        return ToScriptElement(declaration=script, imports=imports)


class ClassMethod(Block):
    """
    Handles static method as well
    """

    def __init__(self, method_type: ClassMethodType[Type], name: str = ''):
        self.method_type = method_type
        inputs = []

        self.method = method_type.get_method()
        inputs = set_inputs_from_function(self.method, inputs)

        self.argument_names = [i.name for i in inputs]
        output_name = f"method result of {method_type.name}"
        output = output_from_function(function=self.method, name=output_name)
        Block.__init__(self, inputs, [output], name=name)

    def equivalent_hash(self):
        classname = self.method_type.class_.__name__
        return len(classname) + 7 * len(self.method_type.name)

    def equivalent(self, other: 'ClassMethod'):
        classname = self.method_type.class_.__name__
        other_classname = other.method_type.class_.__name__
        same_class = classname == other_classname
        same_method = self.method_type.name == other.method_type.name
        return Block.equivalent(self, other) and same_class and same_method

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        classname = full_classname(object_=self.method_type.class_, compute_for='class')
        method_type_dict = {'class_': classname, 'name': self.method_type.name}
        dict_.update({'method_type': method_type_dict})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#') -> 'ClassMethod':
        if 'method_type' in dict_:
            classname = dict_['method_type']['class_']
            method_name = dict_['method_type']['name']
        else:
            # Retro-compatibility
            classname = dict_['model_class']
            method_name = dict_['method_name']
        class_ = get_python_class_from_class_name(classname)
        name = dict_['name']
        method_type = ClassMethodType(class_=class_, name=method_name)
        return cls(method_type=method_type, name=name)

    def evaluate(self, values):
        args = {arg_name: values[var] for arg_name, var in zip(self.argument_names, self.inputs) if var in values}
        return [self.method(**args)]

    def _docstring(self):
        method = self.method_type.get_method()
        docstring = method.__doc__
        annotations = get_type_hints(method)
        parsed_docstring = parse_docstring(docstring=docstring, annotations=annotations)
        parsed_attributes = parsed_docstring["attributes"]
        block_docstring = {i: parsed_attributes[i.name] if i.name in parsed_attributes
                           else EMPTY_PARSED_ATTRIBUTE for i in self.inputs}
        return block_docstring

    def _to_script(self) -> ToScriptElement:
        script = f"ClassMethod(method_type=ClassMethodType(" \
                 f"{self.method_type.class_.__name__}, '{self.method_type.name}')" \
                 f", name='{self.name}')"

        imports = [
            full_classname(object_=self.method_type, compute_for='instance'),
            full_classname(object_=self.method_type.class_, compute_for='class'),
            self.full_classname
        ]

        return ToScriptElement(declaration=script, imports=imports)


class ModelMethod(Block):
    """
    :param method_type: Represent class and method used.
    :type method_type: MethodType[T]
    :param name: Name of the block.
    :type name: str
    """

    def __init__(self, method_type: MethodType[Type], name: str = ''):
        self.method_type = method_type
        inputs = [TypedVariable(type_=method_type.class_, name='model at input')]
        method = method_type.get_method()
        inputs = set_inputs_from_function(method, inputs)

        # Storing argument names
        self.argument_names = [i.name for i in inputs[1:]]

        return_output_name = f"method result of {method_type.name}"
        return_output = output_from_function(function=method, name=return_output_name)

        model_output_name = f"model at output {method_type.name}"
        model_output = TypedVariable(type_=method_type.class_, name=model_output_name)
        outputs = [return_output, model_output]
        if name == "":
            name = f"Model method: {method_type.name}"
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        classname = self.method_type.class_.__name__
        return len(classname) + 7 * len(self.method_type.name)

    def equivalent(self, other: 'ModelMethod'):
        classname = self.method_type.class_.__name__
        other_classname = other.method_type.class_.__name__
        same_model = classname == other_classname
        same_method = self.method_type.name == other.method_type.name
        return Block.equivalent(self, other) and same_model and same_method

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        classname = full_classname(object_=self.method_type.class_, compute_for='class')
        method_type_dict = {'class_': classname, 'name': self.method_type.name}
        dict_.update({'method_type': method_type_dict})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#') -> 'ModelMethod':
        if 'method_type' in dict_:
            classname = dict_['method_type']['class_']
            method_name = dict_['method_type']['name']
        else:
            # Retro-compatibility
            classname = dict_['model_class']
            method_name = dict_['method_name']
        class_ = get_python_class_from_class_name(classname)
        name = dict_['name']
        method_type = MethodType(class_=class_, name=method_name)
        return cls(method_type=method_type, name=name)

    def evaluate(self, values):
        args = {arg_name: values[var] for arg_name, var in zip(self.argument_names, self.inputs[1:]) if var in values}
        return [getattr(values[self.inputs[0]], self.method_type.name)(**args), values[self.inputs[0]]]

    def package_mix(self):
        return {self.method_type.class_.__module__.split('.')[0]: 1}

    def _docstring(self):
        method = self.method_type.get_method()
        docstring = method.__doc__
        annotations = get_type_hints(method)
        parsed_docstring = parse_docstring(docstring=docstring, annotations=annotations)
        parsed_attributes = parsed_docstring["attributes"]
        block_docstring = {i: parsed_attributes[i.name] if i.name in parsed_attributes
                           else EMPTY_PARSED_ATTRIBUTE for i in self.inputs}
        return block_docstring

    def _to_script(self) -> ToScriptElement:
        script = f"ModelMethod(method_type=MethodType(" \
                 f"{self.method_type.class_.__name__}, '{self.method_type.name}')" \
                 f", name='{self.name}')"

        imports = [
            full_classname(object_=self.method_type, compute_for='instance'),
            full_classname(object_=self.method_type.class_, compute_for='class'),
            self.full_classname
        ]
        return ToScriptElement(declaration=script, imports=imports)


class Sequence(Block):
    def __init__(self, number_arguments: int, name: str = ''):
        self.number_arguments = number_arguments
        inputs = [Variable(name=f"Sequence element {i}") for i in range(self.number_arguments)]
        outputs = [TypedVariable(type_=list, name='sequence')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return self.number_arguments

    def equivalent(self, other):
        return Block.equivalent(self, other) and self.number_arguments == other.number_arguments

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_['number_arguments'] = self.number_arguments
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        return cls(dict_['number_arguments'], dict_['name'])

    def evaluate(self, values):
        return [[values[var] for var in self.inputs]]

    def _to_script(self) -> ToScriptElement:
        script = f"Sequence(number_arguments={len(self.inputs)}, name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Concatenate(Block):
    def __init__(self, number_arguments: int, name: str = ''):
        self.number_arguments = number_arguments
        inputs = [Variable(name=f"Sequence element {i}") for i in range(self.number_arguments)]
        outputs = [TypedVariable(type_=list, name='sequence')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return self.number_arguments

    def equivalent(self, other):
        return Block.equivalent(self, other) and self.number_arguments == other.number_arguments

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_['number_arguments'] = self.number_arguments
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        return cls(dict_['number_arguments'], dict_['name'])

    def evaluate(self, values: Dict[Variable, Any]):
        list_values = list(values.values())
        return [concatenate(list_values)]

    def _to_script(self) -> ToScriptElement:
        script = f"Concatenate(number_arguments={len(self.inputs)}, name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class WorkflowBlock(Block):
    """
    Wrapper around workflow to put it in a block of another workflow
    Even if a workflow is a block, it can't be used directly as it has
    a different behavior
    than a Block in eq and hash which is problematic to handle in dicts
    for example

    :param workflow: The WorkflowBlock's workflow
    :type workflow: Workflow
    """

    def __init__(self, workflow: Workflow, name: str = ''):
        self.workflow = workflow
        # TODO: configuring port internal connections
        self.input_connections = None
        self.output_connections = None
        inputs = []
        for variable in self.workflow.inputs:
            input_ = variable.copy()
            input_.name = f"{name} - {variable.name}"
            inputs.append(input_)

        outputs = [self.workflow.output.copy()]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return hash(self.workflow)

    def equivalent(self, other):
        if not Block.equivalent(self, other):
            return False
        return self.workflow == other.workflow

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_.update({'workflow': self.workflow.to_dict(use_pointers=use_pointers, memo=memo, path=f'{path}/workflow')})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#') -> 'WorkflowBlock':
        workflow = Workflow.dict_to_object(dict_=dict_["workflow"])
        # , force_generic = force_generic,
        # global_dict = global_dict, pointers_memo = pointers_memo,
        # path = f"{path}/workflow"
        return cls(workflow=workflow, name=dict_['name'])

    def evaluate(self, values):
        args = {self.inputs.index(input_): v for input_, v in values.items()}
        workflow_run = self.workflow.run(args)
        return [workflow_run.output_value]

    def package_mix(self):
        return self.workflow.package_mix()

    def _docstring(self):
        workflow_docstrings = self.workflow._docstring()
        docstring = {}
        for block_docstring in workflow_docstrings:
            for input_ in self.workflow.inputs:
                if block_docstring and input_ in block_docstring:
                    docstring[input_] = block_docstring[input_]
        return docstring

    def _to_script(self) -> ToScriptElement:
        prefix = 'sub_'
        workflow_script = self.workflow._to_script(prefix)
        script_workflow = f"\n# --- Subworkflow --- \n" \
            f"{workflow_script.declaration}" \
            f"# --- End Subworkflow --- \n"

        script = f"WorkflowBlock(workflow={prefix}workflow, name='{self.name}')"

        imports = workflow_script.imports + [self.full_classname]
        return ToScriptElement(declaration=script, before_declaration=script_workflow, imports=imports)


class ForEach(Block):
    """
    A block to iterate on an input and perform an parralel for (iterations are not dependant)
    :param workflow_block: The WorkflowBlock on which iterate.
    :type workflow_block: WorkflowBlock
    :param iter_input_index: Index of iterable input in worklow_block.inputs
    :type iter_input_index: int
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, workflow_block: 'WorkflowBlock', iter_input_index: int, name: str = ''):
        self.workflow_block = workflow_block
        self.iter_input_index = iter_input_index
        self.iter_input = self.workflow_block.inputs[iter_input_index]
        inputs = []
        for i, workflow_input in enumerate(self.workflow_block.inputs):
            if i == iter_input_index:
                variable_name = 'Iterable input: ' + workflow_input.name
                inputs.append(Variable(name=variable_name))
            else:
                input_ = workflow_input.copy()
                input_.name = 'binding ' + input_.name
                inputs.append(input_)
        output_variable = Variable(name='Foreach output')
        self.output_connections = None  # TODO: configuring port internal connections
        self.input_connections = None
        Block.__init__(self, inputs, [output_variable], name=name)

    def equivalent_hash(self):
        wb_hash = int(self.workflow_block.equivalent_hash() % 10e5)
        return wb_hash + self.iter_input_index

    def equivalent(self, other):
        input_eq = self.iter_input_index == other.iter_input_index
        wb_eq = self.workflow_block.equivalent(other.workflow_block)
        return Block.equivalent(self, other) and wb_eq and input_eq

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        wb_dict = self.workflow_block.to_dict(use_pointers=use_pointers, memo=memo, path=f"{path}/worklow_block")
        dict_.update({'workflow_block': wb_dict, 'iter_input_index': self.iter_input_index})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        workflow_block = WorkflowBlock.dict_to_object(dict_=dict_['workflow_block'])
        # , force_generic = force_generic,
        # global_dict = global_dict, pointers_memo = pointers_memo,
        # path = f"{path}/workflow_block"
        return cls(workflow_block=workflow_block, iter_input_index=dict_['iter_input_index'], name=dict_['name'])

    def evaluate(self, values):
        values_workflow = {var2: values[var1] for var1, var2 in zip(self.inputs, self.workflow_block.inputs)}
        output_values = []
        for value in values_workflow[self.iter_input]:
            values_workflow[self.iter_input] = value
            output = self.workflow_block.evaluate(values_workflow)[0]
            output_values.append(output)
        return [output_values]

    def _docstring(self):
        wb_docstring = self.workflow_block._docstring()
        block_docstring = {}
        for input_, workflow_input in zip(self.inputs, self.workflow_block.workflow.inputs):
            block_docstring[input_] = wb_docstring[workflow_input]
        return block_docstring

    def _to_script(self) -> ToScriptElement:
        wfblock_script_elements = self.workflow_block._to_script()
        wfblock_script = f"{wfblock_script_elements.before_declaration}\n" \
                         f"wfblock = {wfblock_script_elements.declaration}"
        foreach_script = f"ForEach(workflow_block=wfblock, iter_input_index={self.iter_input_index})"

        imports = wfblock_script_elements.imports + [self.full_classname]
        return ToScriptElement(declaration=foreach_script, before_declaration=wfblock_script, imports=imports)


class Unpacker(Block):
    def __init__(self, indices: List[int], name: str = ''):
        self.indices = indices
        outputs = [Variable(name=f"output_{i}") for i in indices]
        Block.__init__(self, inputs=[Variable(name="input_sequence")], outputs=outputs, name=name)

    def equivalent(self, other):
        return Block.equivalent(self, other) and self.indices == other.indices

    def equivalent_hash(self):
        return len(self.indices)

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_['indices'] = self.indices
        return dict_

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#') -> 'Unpacker':
        return cls(dict_['indices'], dict_['name'])

    def evaluate(self, values):
        return [values[self.inputs[0]][i] for i in self.indices]

    def _to_script(self) -> ToScriptElement:
        script = f"Unpacker(indices={self.indices}, name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Flatten(Block):
    def __init__(self, name: str = ''):
        inputs = [Variable(name='input_sequence')]
        outputs = [Variable(name='flatten_sequence')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return 1

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#') -> 'Flatten':
        return cls(dict_['name'])

    def evaluate(self, values):
        output = []
        for value in values[self.inputs[0]]:
            output.extend(value)
        return [output]

    def _to_script(self) -> ToScriptElement:
        script = f"Flatten(name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Product(Block):
    def __init__(self, number_list: int, name: str = ''):
        self.number_list = number_list
        inputs = [Variable(name='list_product_' + str(i)) for i in range(self.number_list)]
        output_variable = Variable(name='Product output')
        Block.__init__(self, inputs, [output_variable], name=name)

    def equivalent_hash(self):
        return self.number_list

    def equivalent(self, other):
        return Block.equivalent(self, other) and self.number_list == other.number_list

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_['number_list'] = self.number_list
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        number_list = dict_['number_list']
        return cls(number_list=number_list, name=dict_['name'])

    def evaluate(self, values):
        """
        Computes the block: use itertools.product
        """
        list_product = [values[var] for var in self.inputs]
        output_value = list(itertools.product(*list_product))
        return [output_value]

    def _to_script(self) -> ToScriptElement:
        script = f"Product(number_list={self.number_list}, name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Filter(Block):
    """
    :param filters: A list of dictionaries, each corresponding to a value to filter.
                    The dictionary should be as follows :
                    *{'attribute' : Name of attribute to filter (str),
                      'comparison_operator' : choose between gt, lt, get, let (standing for greater than, lower than,
                                                                    geater or equal than, lower or equal than) (str),
                      'bound' :  the value (float)}*
    :type filters: list[DessiaFilter]
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, filters: List[DessiaFilter], logical_operator: str = "and", name: str = ''):
        self.filters = filters
        self.logical_operator = logical_operator
        inputs = [Variable(name='input_list')]
        outputs = [Variable(name='output_list')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent(self, other):
        return Block.equivalent(self, other) and self.filters == other.filters

    def equivalent_hash(self):
        hashes = [hash(f) for f in self.filters]
        return int(sum(hashes) % 10e5)

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        filters_dict = [f.to_dict(use_pointers=use_pointers, memo=memo, path=f"{path}/filters/{i}")
                        for i, f in enumerate(self.filters)]
        dict_.update({"filters": filters_dict, "logical_operator": self.logical_operator})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        filters = [DessiaFilter.dict_to_object(dict_=d, force_generic=force_generic, global_dict=global_dict,
                                               pointers_memo=pointers_memo, path=f"{path}/filters/{i}")
                   for i, d in enumerate(dict_['filters'])]
        return cls(filters=filters, logical_operator=dict_["logical_operator"], name=dict_["name"])

    def evaluate(self, values):
        filters_list = FiltersList(self.filters, self.logical_operator)
        return [filters_list.apply(values[self.inputs[0]])]

    def _to_script(self) -> ToScriptElement:
        filter_variables = [f"DessiaFilter("
                            f"attribute='{f.attribute}', comparison_operator='{f.comparison_operator}', "
                            f"bound={f.bound}, name='{f.name}')" for f in self.filters]
        filters = '[' + ",".join(filter_variables) + ']'
        script = f"Filter(filters={filters}, logical_operator='{self.logical_operator}', name='{self.name}')"

        imports = [DessiaFilter("", "", 0).full_classname, self.full_classname]
        return ToScriptElement(declaration=script, imports=imports)


class Display(Block):
    _displayable_input = 0
    _non_editable_attributes = ['inputs']

    def __init__(self, inputs: List[Variable] = None, order: int = None, name: str = ''):
        """
        Abstract class for display behaviors
        """
        if order is not None:
            warnings.warn("Display Block : order argument is deprecated and will be removed in a future version."
                          "You can safely remove it from your block definition", DeprecationWarning)
        self.order = order
        if inputs is None:
            self.warn_deprecation()
            inputs = [TypedVariable(type_=DessiaObject, name='Model to Display')]
        output = TypedVariable(type_=DisplayObject, name="Display Object")
        Block.__init__(self, inputs=inputs, outputs=[output], name=name)

        self._type = None
        self._selector = None

    @staticmethod
    def warn_deprecation():
        warnings.warn("Display Block used as a generator for the displays of an object is deprecated."
                      "ts display behavior will be faulty. Please use the specific block"
                      "to generate wanted displays (MultiPlot, CadView, PlotData, Markdown)", DeprecationWarning)

    @property
    def type_(self):
        if self._type:
            return self._type
        if self.__class__ is Display:
            self.warn_deprecation()
            return ""
        raise NotImplementedError(f"type_ attribute is not implemented for block of type '{type(self)}'")

    @property
    def selector(self):
        if self._selector:
            return self._selector
        if self.__class__ is Display:
            self.warn_deprecation()
            return ""
        raise NotImplementedError(f"selector attribute is not implemented for block of type '{type(self)}'")

    def _display_settings(self, block_index: int) -> DisplaySetting:
        args = {'block_index': block_index}
        return DisplaySetting(selector=None, type_=self.type_, method="block_display", serialize_data=True,
                              arguments=args)

    def evaluate(self, values):
        object_ = values[self.inputs[0]]
        settings = object_._display_settings_from_selector(self.selector)
        data = attrmethod_getter(object_, settings.method)()
        return [DisplayObject(type_=settings.type, data=serialize(data), name=self.name)]

    def _to_script(self) -> ToScriptElement:
        script = f"{self.__class__.__name__}(name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class MultiPlot(Display):
    """
    Generates a PlotData multiplot which axes will be the given attributes
    (can be deep attributes with the '/' separator)

    :param attributes: A List of all attributes that will be shown on axes in the ParallelPlot window.
    :type attributes: List[str]
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, attributes: List[str], order: int = None, name: str = ''):
        if order is not None:
            warnings.warn("Display Block : order argument is deprecated and will be removed in a future version."
                          "You can safely remove it from your block definition", DeprecationWarning)
        self.order = order
        self.attributes = attributes
        Display.__init__(self, inputs=[TypedVariable(List[DessiaObject])], name=name)
        self.inputs[0].name = 'Input List'
        self._type = "plot_data"
        self._selector = None

    def equivalent(self, other):
        same_attributes = self.attributes == other.attributes
        return Block.equivalent(self, other) and same_attributes

    def equivalent_hash(self):
        return sum(len(a) for a in self.attributes)

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_['attributes'] = self.attributes
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        return cls(attributes=dict_['attributes'], name=dict_['name'])

    def evaluate(self, values):
        import plot_data
        objects = values[self.inputs[self._displayable_input]]
        attr_values = [{a: enhanced_deep_attr(o, a) for a in self.attributes} for o in objects]
        values2d = [{key: val[key]} for key in self.attributes[:2] for val in attr_values]
        tooltip = plot_data.Tooltip(name='Tooltip', attributes=self.attributes)

        scatterplot = plot_data.Scatter(tooltip=tooltip, x_variable=self.attributes[0], y_variable=self.attributes[1],
                                        elements=values2d, name='Scatter Plot')

        parallelplot = plot_data.ParallelPlot(disposition='horizontal', axes=self.attributes,
                                              rgbs=[(192, 11, 11), (14, 192, 11), (11, 11, 192)], elements=attr_values)
        plots = [scatterplot, parallelplot]
        sizes = [plot_data.Window(width=560, height=300), plot_data.Window(width=560, height=300)]
        multiplot = plot_data.MultiplePlots(elements=attr_values, plots=plots, sizes=sizes,
                                            coords=[(0, 0), (0, 300)], name='Results plot')
        return [DisplayObject(type_=self.type_, data=[multiplot.to_dict()])]

    def _to_script(self) -> ToScriptElement:
        script = f"MultiPlot(attributes={self.attributes}, name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class CadView(Display):
    """
    Generates a DisplayObject that is displayable in 3D Viewer features (BabylonJS, ...)

    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, name: str = ''):
        input_ = TypedVariable(DessiaObject, name="Model to display")
        Display.__init__(self, inputs=[input_], name=name)

        self._type = "babylon_data"
        self._selector = "cad"


class Markdown(Display):
    """
    Generates the markdown representation of an object

    :param name: Name of the block.
    :type name: str
    """
    def __init__(self, name: str = ''):
        input_ = TypedVariable(DessiaObject, name="Model to display")
        Display.__init__(self, inputs=[input_], name=name)

        self._type = "markdown"
        self._selector = "markdown"


class PlotData(Display):
    """
    Generates a DisplayObject that is displayable in PlotData features.
    Uses the the input object's plot_data method.

    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, name: str = ''):
        input_ = TypedVariable(DessiaObject, name="Model to display")
        Display.__init__(self, inputs=[input_], name=name)

        self._type = "plot_data"
        self._selector = "plot_data"


class ModelAttribute(Block):
    """
    :param attribute_name: The name of the attribute to select.
    :type attribute_name: str
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, attribute_name: str, name: str = ''):
        self.attribute_name = attribute_name
        inputs = [Variable(name='Model')]
        outputs = [Variable(name='Model attribute')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return len(self.attribute_name)

    def equivalent(self, other):
        return Block.equivalent(self, other) and self.attribute_name == other.attribute_name

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_.update({'attribute_name': self.attribute_name})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        return cls(dict_['attribute_name'], dict_['name'])

    def evaluate(self, values):
        return [enhanced_deep_attr(values[self.inputs[0]], self.attribute_name)]

    def _to_script(self) -> ToScriptElement:
        script = f"ModelAttribute(attribute_name='{self.attribute_name}', name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class SetModelAttribute(Block):
    """
    :param attribute_name: The name of the attribute to set.
    :type attribute_name: str
    :param name: The name of the block.
    :type name: str
    """

    def __init__(self, attribute_name: str, name: str = ''):
        self.attribute_name = attribute_name
        inputs = [Variable(name='Model'), Variable(name=f'Value to insert for attribute {attribute_name}')]
        outputs = [Variable(name=f'Model with changed attribute {attribute_name}')]
        Block.__init__(self, inputs, outputs, name=name)

    def equivalent_hash(self):
        return 3 + len(self.attribute_name)

    def equivalent(self, other):
        return Block.equivalent(self, other) and self.attribute_name == other.attribute_name

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_.update({'attribute_name': self.attribute_name})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        return cls(dict_['attribute_name'], dict_['name'])

    def evaluate(self, values):
        model = values[self.inputs[0]]
        setattr(model, self.attribute_name, values[self.inputs[1]])
        return [model]

    def _to_script(self) -> ToScriptElement:
        script = f"SetModelAttribute(attribute_name='{self.attribute_name}', name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Sum(Block):
    def __init__(self, number_elements: int = 2, name: str = ''):
        self.number_elements = number_elements
        inputs = [Variable(name=f"Sum element {i + 1}") for i in range(number_elements)]
        Block.__init__(self, inputs=inputs, outputs=[Variable(name='Sum')], name=name)

    def equivalent_hash(self):
        return self.number_elements

    def equivalent(self, other):
        return Block.equivalent(self, other) and self.number_elements == other.number_elements

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_.update({'number_elements': self.number_elements})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        return cls(dict_['number_elements'], dict_['name'])

    @staticmethod
    def evaluate(values):
        return [sum(values)]

    def _to_script(self) -> ToScriptElement:
        script = f"Sum(number_elements={self.number_elements}, name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Substraction(Block):
    def __init__(self, name: str = ''):
        Block.__init__(self, [Variable(name='+'), Variable(name='-')], [Variable(name='Substraction')], name=name)

    def evaluate(self, values):
        return [values[self.inputs[0]] - values[self.inputs[1]]]

    def _to_script(self) -> ToScriptElement:
        script = f"Substraction(name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class ConcatenateStrings(Block):
    """
    Concatenates the n input elements, separate by the separator input, into one string

    :param number_elements: Number of block inputs
    :type number_elements: int
    :param separator: Character used to joins the input elements together
    :type separator: str
    """
    def __init__(self, number_elements: int = 2, separator: str = "", name: str = ''):
        self.number_elements = number_elements
        self.separator = separator
        inputs = [TypedVariableWithDefaultValue(name=f"Substring {i + 1}", type_=str, default_value="")
                  for i in range(number_elements)]
        output = TypedVariable(name="Concatenation", type_=str)
        Block.__init__(self, inputs=inputs, outputs=[output], name=name)

    def equivalent_hash(self):
        return self.number_elements + hash(self.separator)

    def equivalent(self, other):
        same_number = self.number_elements == other.number_elements
        same_separator = self.separator == other.separator
        return Block.equivalent(self, other) and same_number and same_separator

    def to_dict(self, use_pointers=True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_.update({'number_elements': self.number_elements, "separator": self.separator})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        return cls(number_elements=dict_['number_elements'], separator=dict_["separator"], name=dict_['name'])

    def evaluate(self, values):
        chunks = [values[i] for i in self.inputs]
        return [self.separator.join(chunks)]

    def _to_script(self) -> ToScriptElement:
        script = f"ConcatenateStrings(number_elements={self.number_elements}, " \
                 f"separator='{self.separator}', " \
                 f"name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])


class Export(Block):
    def __init__(self, method_type: MethodType[Type],  text: bool, extension: str,
                 filename: str = "export", name: str = ""):
        """
        Block that enables an export of an object calling its configured method.
        Only Methods that yields streams (and not files) should be used.

        The file generated will be called {filename}.{extension}

        :param method_type: An object that have a class_ input (which is the class of the incoming model)
            and a name (which is the name of the method that will be called).
        :type method_type: MethodType[T]
        :param text: Whether the export is of type text or not
        :type text: bool
        :param extension: Extension of the resulting file (ex: json or xlsx)
        :type extension: str
        :param filename: Name of the resulting file without its extension
        :type filename: str
        :param name: Name of the block.
        :type name: str
        """
        self.method_type = method_type
        if not filename:
            filename = "export"
        self.filename = filename

        method = method_type.get_method()

        self.extension = extension
        self.text = text

        output = output_from_function(function=method, name="export_output")
        inputs = [TypedVariable(type_=method_type.class_, name="model_to_export"),
                  TypedVariableWithDefaultValue(type_=str, default_value=filename, name="filename")]
        Block.__init__(self, inputs=inputs, outputs=[output], name=name)

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        classname = full_classname(object_=self.method_type.class_, compute_for='class')
        method_type_dict = {'class_': classname, 'name': self.method_type.name}
        dict_.update({"method_type": method_type_dict, "extension": self.extension,
                      "text": self.text, "filename": self.filename})
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#') -> 'Export':
        class_ = get_python_class_from_class_name(dict_['method_type']['class_'])
        method_type = MethodType(class_=class_, name=dict_['method_type']['name'])
        if "export_name" in dict_:
            # RetroCompat
            filename = dict_["export_name"]
        else:
            filename = dict_["filename"]
        return cls(method_type=method_type, text=dict_['text'], filename=filename,
                   extension=dict_["extension"], name=dict_["name"])

    def evaluate(self, values):
        filename = f"{values.pop(self.inputs[-1])}.{self.extension}"
        if self.text:
            stream = StringFile(filename)
        else:
            stream = BinaryFile(filename)
        getattr(values[self.inputs[0]], self.method_type.name)(stream)
        return [stream]

    def _export_format(self, block_index: int):
        args = {"block_index": block_index}
        return {"extension": self.extension, "method_name": "export", "text": self.text,
                "export_name": self.filename, "args": args}

    def _to_script(self) -> ToScriptElement:
        script = f"Export(method_type=MethodType(" \
                 f"{full_classname(object_=self.method_type.class_, compute_for='class')}, '{self.method_type.name}')" \
                 f", filename='{self.filename}', extension='{self.extension}'" \
                 f", text={self.text}, name='{self.name}')"

        imports = [self.full_classname, full_classname(object_=self.method_type, compute_for='instance'),
                   full_classname(object_=self.method_type.class_, compute_for='class')]
        return ToScriptElement(declaration=script, imports=imports)


class Archive(Block):
    def __init__(self, number_exports: int = 1, filename: str = "archive", name: str = ""):
        """
        A block that takes n inputs and store them in a archive (ZIP,...)

        :param number_exports: The number of files that will be stored in the archive
        :type number_exports: int
        :param filename: Name of the resulting archive file without its extension
        :type filename: str
        :param name: Name of the block.
        :type name: str
        """
        self.number_exports = number_exports
        self.filename = filename
        inputs = [Variable(name="export_" + str(i)) for i in range(number_exports)]
        inputs.append(TypedVariableWithDefaultValue(type_=str, default_value=filename, name="filename"))
        Block.__init__(self, inputs=inputs, outputs=[Variable(name="zip archive")], name=name)

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#'):
        dict_ = Block.to_dict(self, use_pointers=use_pointers, memo=memo, path=path)
        dict_['number_exports'] = len(self.inputs) - 1   # Filename is also a block input
        dict_["filename"] = self.filename
        return dict_

    @classmethod
    @set_block_variable_names_from_dict
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False,
                       global_dict=None, pointers_memo: Dict[str, Any] = None, path: str = '#'):
        if "export_name" in dict_:
            # RetroCompat
            filename = dict_["export_name"]
        else:
            filename = dict_["filename"]
        return cls(number_exports=dict_["number_exports"], filename=filename, name=dict_['name'])

    def evaluate(self, values):
        name_input = self.inputs[-1]
        archive_name = f"{values.pop(name_input)}.zip"
        archive = BinaryFile(archive_name)
        with ZipFile(archive, 'w') as zip_archive:
            for input_ in self.inputs[:-1]:  # Filename is last block input
                value = values[input_]
                if isinstance(value, StringFile):
                    with zip_archive.open(value.filename, 'w') as file:
                        file.write(value.getvalue().encode('utf-8'))
                elif isinstance(value, BinaryFile):
                    with zip_archive.open(value.filename, 'w') as file:
                        file.write(value.getbuffer())
                else:
                    raise ValueError(f"Archive input is not a file-like object. Got '{value}' of type {type(value)}")
        return [archive]

    def _export_format(self, block_index: int):
        return {"extension": "zip", "method_name": "export", "text": False,
                "export_name": self.filename, "args": {"block_index": block_index}}

    def _to_script(self) -> ToScriptElement:
        script = f"Archive(number_exports={self.number_exports}, filename='{self.filename}', name='{self.name}')"
        return ToScriptElement(declaration=script, imports=[self.full_classname])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Interpolation tools

"""

from typing import List


def istep_from_value_on_list(list_: List[float], value: float,
                             extrapolate=False):
    """
    Return the float index of a value in a list of objects
    """
    for ipoint, (point1, point2) in enumerate(zip(list_[:-1],
                                                  list_[1:])):

        point1_s, point2_s = sorted((point1, point2))
        if point1_s <= value <= point2_s:
            alpha = (value - point1_s) / (point2_s - point1_s)
            if alpha < 0 or alpha > 1:
                raise ValueError
            return ipoint + alpha
    # values = [p for p in list_]

    if extrapolate:
        if abs(list_[0] - value) < abs(list_[-1] - value):
            # Closer to left
            return 0.
        # else:
            # Closer to right
        return len(list_) - 1

    min_values = min(list_)
    max_values = max(list_)
    raise ValueError(f'Specified value not found in list_: {value} not in [{min_values}, {max_values}]')


def interpolate_from_istep(objects, istep: float):
    """
    Return the interpolated object from a float index and a list of objects
    """
    n_objects = len(objects)
    if (istep < 0) or (istep > n_objects - 1):
        raise ValueError('Extrapolating is not supported')
    istep1 = int(istep)
    if istep1 == istep:
        # No interpolation needed
        return objects[int(istep)]
    # else:
    alpha = istep - istep1
    point1 = objects[istep1]
    point2 = objects[istep1 + 1]
    return (1 - alpha) * point1 + (alpha) * point2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 12 19:24:11 2022

@author: steven
"""

from typing import Dict, Any, Tuple
try:
    from typing import TypedDict  # >=3.8
except ImportError:
    from mypy_extensions import TypedDict  # <=3.7


class ParsedAttribute(TypedDict):
    desc: str
    type_: str
    annotation: str


class ParsedDocstring(TypedDict):
    description: str
    attributes: Dict[str, ParsedAttribute]


def parse_docstring(docstring: str, annotations: Dict[str, Any]) -> ParsedDocstring:
    """
    Parse docstring of given class. Refer to docs to see how docstrings
    should be built.
    """
    if docstring:
        splitted_docstring = docstring.split(':param ')
        parsed_docstring = {"description": splitted_docstring[0].strip()}
        params = splitted_docstring[1:]
        args = {}
        for param in params:
            argname, parsed_attribute = parse_attribute(param, annotations)
            args[argname] = parsed_attribute
            # TODO Should be serialize typing ?
        parsed_docstring.update({'attributes': args})
        return parsed_docstring
    return {'description': "", 'attributes': {}}


def parse_attribute(param, annotations) -> Tuple[str, ParsedAttribute]:
    splitted_param = param.split(':type ')
    arg = splitted_param[0]
    typestr = splitted_param[1]
    argname, argdesc = arg.split(":", maxsplit=1)
    argtype = typestr.split(argname + ":")[-1]
    annotation = annotations[argname]
    parsed_attribute = {'desc': argdesc.strip(),
                        'type_': argtype.strip(),
                        'annotation': str(annotation)}
    return argname, parsed_attribute


EMPTY_PARSED_ATTRIBUTE = {"desc": "", "type": "", "annotation": ""}
FAILED_DOCSTRING_PARSING = {'description': 'Docstring parsing failed', 'attributes': {}}
FAILED_ATTRIBUTE_PARSING = {"desc": 'Attribute documentation parsing failed',
                            "type": "", "annotation": ""}



from dessia_common.workflow import Workflow, Pipe
from dessia_common.workflow.blocks import ClassMethod, PlotData, CadView, Markdown
from dessia_common.typings import ClassMethodType
from dessia_common.forms import StandaloneObject

cmt = ClassMethodType(class_=StandaloneObject, name="generate")
cmb = ClassMethod(method_type=cmt, name="Generator")

cadview = CadView("3D")
plotdata = PlotData("2D")
markdown = Markdown("MD")

blocks = [cmb, cadview, plotdata, markdown]
pipes = [Pipe(input_variable=cmb.outputs[0], output_variable=cadview.inputs[0]),
         Pipe(input_variable=cmb.outputs[0], output_variable=plotdata.inputs[0]),
         Pipe(input_variable=cmb.outputs[0], output_variable=markdown.inputs[0])]

workflow = Workflow(blocks=blocks, pipes=pipes, output=cmb.outputs[0])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Tests for cars data
"""
import pkg_resources
from dessia_common.vectored_objects import Catalog, ParetoSettings, Objective, ObjectiveSettings
from dessia_common import DessiaFilter
from dessia_common.tests import Car, CarWithFeatures
from dessia_common.files import StringFile

choice_args = ['MPG', 'Cylinders', 'Displacement', 'Horsepower', 'Weight', 'Acceleration', 'Model']  # Ordered

minimized_attributes = {'MPG': False, 'Horsepower': True, 'Weight': True, 'Acceleration': False}
coefficients = {'Cylinders': 0, 'MPG': -0.70, 'Displacement': 0,
                'Horsepower': 0, 'Weight': 0.70, 'Acceleration': 0, 'Model': 0}

pareto_settings = ParetoSettings(minimized_attributes=minimized_attributes, enabled=True)

csv_cars = pkg_resources.resource_stream('dessia_common', 'models/data/cars.csv')
stream = StringFile.from_stream(csv_cars)

csv_cars.seek(0)
catalog = Catalog.from_csv(csv_cars)
catalog.name = 'Cars dataset'
catalog.pareto_settings = pareto_settings

filters = [DessiaFilter(attribute="Weight", comparison_operator="lt", bound=4000)]

# Empty objective because this hasn't been used/tested for a while
objective_settings = ObjectiveSettings()
objective = Objective({}, {}, objective_settings)

filtered_catalog = catalog.filter_(filters)
merged_catalog = Catalog.concatenate(catalogs=[catalog, filtered_catalog])

# Used models
all_cars_no_feat = Car.from_csv(stream)
all_cars_wi_feat = CarWithFeatures.from_csv(stream)
